"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Rewrite PPQ APIs.
"""

import os
from typing import Any, Callable, Collection, Dict, Iterable, Optional, Sequence

from onnx import ModelProto

from mppq.api.extension import (
    _PLATFORM_TO_DISPATCHER_,
    _PLATFORM_TO_PARSER_,
    _PLATFORM_TO_QUANTIZER_,
)
from mppq.common import DEFAULT_QUANTIZE_OP
from mppq.dispatcher import DISPATCHER_TABLE
from mppq.dispatcher.base import GraphDispatcher
from mppq.dispatcher.scope import IgnoredScope
from mppq.executor.base import BaseGraphExecutor, GraphInput
from mppq.frontend import EXPORTER, PARSER
from mppq.ir.base.command import GraphCommand, GraphCommandType
from mppq.ir.base.graph import BaseGraph, GraphBuilder, GraphExporter
from mppq.ir.morph import GraphFormatter, GraphMerger, GraphReplacer
from mppq.quant import TargetPrecision
from mppq.quantizer import QUANTIZER, BaseQuantizer
from mppq.quantizer.base import QuantizationOptimizationPass


def load_graph(graph_file: Any, parser: GraphBuilder) -> BaseGraph:
    r"""加载模型并解析为PPQ IR Graph.

    Args:
        graph_file: 模型文件路径，或者已加载的模型对象。
        parser: 解析模型的GraphBuilder对象。

    Returns:
        BaseGraph: 解析后的IR Graph对象。
    """

    graph = parser.build(graph_file)
    graph = format_graph(graph)
    return graph


def load_onnx_graph(onnx_file: str | os.PathLike | ModelProto) -> BaseGraph:
    r"""加载ONNX模型并解析为PPQ IR Graph.

    Args:
        onnx_file (str | ModelProto): ONNX文件路径，或者已加载的ONNX模型对象。

    Returns:
        BaseGraph: 解析后的IR Graph对象。
    """

    parser = PARSER["onnx"]()
    return load_graph(onnx_file, parser)


def export_graph(graph: BaseGraph, f: str | os.PathLike, *, exporter: GraphExporter):
    r"""导出PPQ IR Graph为目标格式。

    Args:
        graph (BaseGraph): 待导出IR Graph对象。
        f (str|os.PathLike): 导出文件路径。
        exporter (GraphExporter): 导出Graph的GraphExporter对象。
    """

    exporter.export(f, graph)


def export_onnx_graph(graph: BaseGraph, f: str | os.PathLike):
    r"""导出PPQ IR Graph为目标格式。

    Args:
        graph (BaseGraph): 待导出IR Graph对象。
        f (str|os.PathLike): 导出文件路径。
        exporter (GraphExporter): 导出Graph的GraphExporter对象。
    """

    exporter = EXPORTER["onnx"]()
    export_graph(graph, f, exporter=exporter)


def format_graph(
    graph: BaseGraph,
    format_constant_input: bool = True,
    fuse_bias_add: bool = True,
    fuse_bn: bool = True,
    replace_bn_to_conv: bool = True,
    remove_identity: bool = True,
    remove_isolated: bool = True,
) -> BaseGraph:
    r"""这个函数对计算图进行预处理工作，其主要内容是将计算图的格式进行统一。
    这个函数将会统一 cast, slice, parameter, constant 算子的格式，
    并且执行有关 batchnorm 的合并工作.

    在 PPQ 中，我们不希望出现 Constant 算子，所有 Constant 输入将被当作 parameter variable 连接到下游算子上
    在 PPQ 中，我们不希望出现 Batchnorm 算子，所有 Batchnorm 将被合并
    在 PPQ 中，我们不希望出现权重共享的算子，所有被共享的权重将被复制分裂成多份
    在 PPQ 中，我们不希望出现孤立算子，所有孤立算子将被移除

    # 读取 Onnx 图时，将图中所有以 Constant 节点作为输入的变量转换为 Parameter Variable
    FORMATTER_FORMAT_CONSTANT_INPUT
    # 读取 Onnx 图时，合并图中的 Bias add 节点(Conv, ConvTranspose, Gemm)
    FORMATTER_FUSE_BIAS_ADD
    # 读取 Onnx 图时，合并图中的 Batchnorm 节点(Conv, ConvTranspose, Gemm)
    FORMATTER_FUSE_BN
    # 读取 Onnx 图时，将单独的 Batchnorm 替换为卷积
    FORMATTER_REPLACE_BN_TO_CONV
    # 读取 Onnx 图时，移除图中的 Identity 节点
    FORMATTER_REMOVE_IDENTITY
    # 读取 Onnx 图时，移除图中的孤立节点
    FORMATTER_REMOVE_ISOLATED
    """

    # do graph level optimization
    formatter = GraphReplacer(GraphFormatter(GraphMerger(graph)))
    if format_constant_input:
        formatter(GraphCommand(GraphCommandType.FORMAT_CONSTANT_INPUT))
    formatter(GraphCommand(GraphCommandType.CONVERT_TO_TENSOR))
    formatter(GraphCommand(GraphCommandType.FORMAT_PARAMETERS))

    if fuse_bias_add:
        formatter(GraphCommand(GraphCommandType.FUSE_BIAS_ADD))

    if fuse_bn:
        formatter(GraphCommand(GraphCommandType.FUSE_BN))

    if replace_bn_to_conv:
        formatter(GraphCommand(GraphCommandType.REPLACE_BATCHNORM_TO_CONV))

    formatter(GraphCommand(GraphCommandType.FORMAT_CAST))
    formatter(GraphCommand(GraphCommandType.FORMAT_SLICE))
    formatter(GraphCommand(GraphCommandType.FORMAT_CLIP))
    formatter(GraphCommand(GraphCommandType.FORMAT_PAD))
    formatter(GraphCommand(GraphCommandType.FORMAT_RESIZE))

    if remove_identity:
        formatter(GraphCommand(GraphCommandType.REMOVE_IDENTITY))

    if remove_isolated:
        formatter(GraphCommand(GraphCommandType.DELETE_ISOLATED))

    return graph


def dispatch_graph(
    graph: BaseGraph,
    quantize_operations: Collection[str] = DEFAULT_QUANTIZE_OP,
    dispatcher: Optional[str | GraphDispatcher] = None,
    dispatching_override: Optional[Dict[str, TargetPrecision]] = None,
    ignored_scope: Optional[list | IgnoredScope] = None,
    quant_precision: TargetPrecision = TargetPrecision.INT8,
    **kwargs,
) -> BaseGraph:
    """Override the graph dispatching of PPQ. This function split graph into pieces.

    Dispatching results are assigned to each operator in the graph.

    Args:
        graph (BaseGraph): loaded onnx graph
        quantize_operations (Set[str]): an operation list that can be quantized.
        dispatcher (str | GraphDispatcher, optional): dispatch algorithm.
            Defaults to "conservative".
        dispatching_override (DispatchingTable, optional): overwrite the dispatching
            results. Defaults to None.

    Returns:
        BaseGraph: graph that each operator is assigned a dispatching target.
    """

    if isinstance(dispatcher, str):
        dispatcher = dispatcher.lower()
        if dispatcher not in DISPATCHER_TABLE:
            raise ValueError(
                f'Can not found dispatcher type "{dispatcher}", check your input again.'
            )
        dispatcher = DISPATCHER_TABLE[dispatcher](graph)
    else:
        if not isinstance(dispatcher, GraphDispatcher):
            raise TypeError(
                'Parameter "dispachter" of function ppq.api.dispatch_graph must be '
                f"str or GraphDispatcher, however {type(dispatcher)} was given."
            )
    assert isinstance(dispatcher, GraphDispatcher)
    if not quantize_operations:
        quantize_operations = DEFAULT_QUANTIZE_OP

    dispatching_table = dispatcher.dispatch(
        graph=graph,
        quant_types=quantize_operations,
        quant_precision=quant_precision,
        fp32_platform=TargetPrecision.FP32,
        soi_platform=TargetPrecision.SOI,
        **kwargs,
    )

    if ignored_scope is not None:
        if isinstance(ignored_scope, list):
            ignored_scope = IgnoredScope(operations=ignored_scope)
        assert isinstance(ignored_scope, IgnoredScope)
        ignored_table = ignored_scope.dispatch(graph)
        dispatching_table.update(ignored_table)

    # override dispatching result
    if dispatching_override is not None:
        if not isinstance(dispatching_override, dict):
            raise TypeError(
                'Parameter "dispatching_override" of function ppq.api.dispatch_graph '
                f"must be a dict, however {type(dispatching_override)} was given."
            )
        for opname, platform in dispatching_override.items():
            if opname not in graph.operations:
                continue
            if not isinstance(platform, int):
                raise TypeError(
                    "Your dispatching table contains a invalid setting of "
                    f"operation {opname}. All platform setting given in dispatching "
                    f"table is expected given as int, however {type(platform)} was "
                    "given."
                )
            dispatching_table[opname] = TargetPrecision(platform)

    for operation in graph.operations.values():
        if operation.name not in dispatching_table:
            raise RuntimeError(
                f"Internal Error, Can not find operation {operation.name} in "
                "dispatching table."
            )
        operation.precision = dispatching_table[operation.name]
    return graph


def load_quantizer(
    model: str | os.PathLike | object,
    platform: int,
    dispatcher: Optional[str] = None,
    dispatching_override: Optional[Dict[str, TargetPrecision]] = None,
    ignored_scope: Optional[list | IgnoredScope] = None,
    quant_precision: TargetPrecision = TargetPrecision.INT8,
    **kwargs,
) -> BaseQuantizer:
    r"""Quantize the graph to target platform.

    Args:
        graph (BaseGraph): loaded onnx graph
        quantize_operations (Set[str]): an operation list that can be quantized.
        dispatcher (str | GraphDispatcher, optional): dispatch algorithm.
            Defaults to "conservative".
        dispatching_override (DispatchingTable, optional): overwrite the dispatching
            results. Defaults to None.
        ignored_scope (list | IgnoredScope, optional): a list of operation names
            that will be ignored during quantization. Defaults to None.
    """

    parser = PARSER[_PLATFORM_TO_PARSER_[platform]]()
    graph = load_graph(model, parser)
    if dispatcher is None:
        dispatcher = _PLATFORM_TO_DISPATCHER_[platform]
    quantizer = QUANTIZER[_PLATFORM_TO_QUANTIZER_[platform]](graph)
    graph = dispatch_graph(
        graph,
        quantizer.quant_operation_types,
        dispatcher,
        dispatching_override,
        ignored_scope,
        quant_precision,
        **kwargs,
    )

    return quantizer


def quantize(
    model: str | os.PathLike,
    f: str | os.PathLike,
    platform: int,
    quant_precision: TargetPrecision = TargetPrecision.INT8,
    example_inputs: Optional[GraphInput] = None,
    calib_dataloader: Optional[Iterable[Any]] = None,
    executor: Optional[BaseGraphExecutor] = None,
    collate_fn: Optional[Callable[[Any], Any]] = None,
    calib_steps: int = 32,
    prequant_settings: Optional[Sequence[dict | QuantizationOptimizationPass]] = None,
    quant_settings: Optional[Sequence[dict | QuantizationOptimizationPass]] = None,
    dispatcher: Optional[str] = None,
    ignored_scope: Optional[list | IgnoredScope] = None,
    **kwargs,
) -> None:
    r"""Quantize the onnx model to target platform and export the quantized model.

    Args:
        model (str | os.PathLike): onnx model path.
        f (str | os.PathLike): output file path.
        platform (int): target platform.
        dispatcher (str | GraphDispatcher, optional): dispatch algorithm.
            Defaults to "conservative".
        ignored_scope (list | IgnoredScope, optional): a list of operation names
            that will be ignored during quantization. Defaults to None.
    """

    parser = PARSER[_PLATFORM_TO_PARSER_[platform]]()
    graph = load_graph(model, parser)
    if dispatcher is None:
        dispatcher = _PLATFORM_TO_DISPATCHER_[platform]
    quantizer = QUANTIZER[_PLATFORM_TO_QUANTIZER_[platform]](graph)
    graph = dispatch_graph(
        graph,
        quantizer.quant_operation_types,
        dispatcher,
        ignored_scope=ignored_scope,
        quant_precision=quant_precision,
        **kwargs,
    )
    quantizer.quantize(
        example_inputs=example_inputs,
        calib_dataloader=calib_dataloader,
        executor=executor,
        collate_fn=collate_fn,
        calib_steps=calib_steps,
        prequant_settings=prequant_settings,
        quant_settings=quant_settings,
    )
    exporter = EXPORTER[_PLATFORM_TO_PARSER_[platform]]()
    export_graph(graph, f, exporter=exporter)



from mppq.dispatcher.aggressive import AggressiveDispatcher
from mppq.dispatcher.allin import AllinDispatcher
from mppq.dispatcher.base import DISPATCHER_TABLE
from mppq.dispatcher.conservative import ConservativeDispatcher
from mppq.dispatcher.perseus import Perseus
from mppq.dispatcher.pointwise import PointDispatcher

__all__ = [
    "AllinDispatcher",
    "DISPATCHER_TABLE",
    "AggressiveDispatcher",
    "ConservativeDispatcher",
    "PointDispatcher",
    "Perseus",
]



from decimal import ROUND_HALF_DOWN, ROUND_HALF_EVEN, ROUND_HALF_UP, Decimal
from math import ceil, floor, log2
from typing import Union

import torch
from torch.autograd import Function

from mppq.quant import RoundingPolicy


# pylint: disable=abstract-method, arguments-differ
class PPQTensorRoundImpl(Function):
    @staticmethod
    def forward(
        ctx,
        value: torch.Tensor,
        policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
    ) -> torch.Tensor:
        """reference: https://en.wikipedia.org/wiki/Rounding"""
        if policy == RoundingPolicy.ROUND_HALF_EVEN:
            # default rounding policy of torch is ROUND_TO_NEAR_EVEN
            # try this: print(torch.Tensor([1.5, 2.5, 3.5, 4.5]).round())
            # However it may generate unexpected results due to version difference.
            return value.round()
        elif policy == RoundingPolicy.ROUND_UP:
            return value.ceil()
        elif policy == RoundingPolicy.ROUND_HALF_TOWARDS_ZERO:
            return torch.sign(value) * torch.ceil(value.abs() - 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO:
            return torch.sign(value) * torch.floor(value.abs() + 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_DOWN:
            return torch.ceil(value - 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_UP:
            return torch.floor(value + 0.5)
        else:
            raise ValueError(f"Unexpected rounding {policy}.")

    @staticmethod
    def backward(ctx, *grads_output: torch.Tensor):
        return grads_output[0], None


def ppq_numerical_round(
    value: float, policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN
) -> int:
    """reference: https://en.wikipedia.org/wiki/Rounding

    decimal definition:
        - decimal.ROUND_CEILING (towards Infinity)
        - decimal.ROUND_DOWN (towards zero)
        - decimal.ROUND_FLOOR (towards -Infinity)
        - decimal.ROUND_HALF_DOWN (to nearest with ties going towards zero)
        - decimal.ROUND_HALF_EVEN (to nearest with ties going to nearest even integer)
        - decimal.ROUND_HALF_UP (to nearest with ties going away from zero)
        - decimal.ROUND_UP (away from zero)
        - decimal.ROUND_05UP (away from zero if last digit after rounding towards zero
          would have been 0 or 5; otherwise towards zero)
    """
    assert isinstance(
        value, float
    ), "numerical round only takes effect on float number."
    if policy == RoundingPolicy.ROUND_HALF_EVEN:
        return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_EVEN))
    elif policy == RoundingPolicy.ROUND_HALF_UP:
        if value > 0:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_UP))
        else:
            return int(
                Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_DOWN)
            )
    elif policy == RoundingPolicy.ROUND_HALF_DOWN:
        if value > 0:
            return int(
                Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_DOWN)
            )
        else:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_UP))
    elif policy == RoundingPolicy.ROUND_HALF_TOWARDS_ZERO:
        return ppq_numerical_round(value, RoundingPolicy.ROUND_HALF_DOWN)
    elif policy == RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO:
        return ppq_numerical_round(value, RoundingPolicy.ROUND_HALF_UP)
    elif policy == RoundingPolicy.ROUND_TO_NEAR_INT:
        if value > 0:
            return floor(value + 0.5)
        else:
            return ceil(value - 0.5)
    elif policy == RoundingPolicy.ROUND_UP:
        return ceil(value)
    else:
        raise ValueError("Unexpected rounding policy found.")


def ppq_tensor_round(
    value: torch.Tensor, policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN
) -> torch.Tensor:
    """reference: https://en.wikipedia.org/wiki/Rounding"""
    round_value = PPQTensorRoundImpl.apply(value, policy)
    assert isinstance(round_value, torch.Tensor)
    return round_value


def ppq_round_to_power_of_2(
    value: Union[float, int], policy: RoundingPolicy = RoundingPolicy.ROUND_UP
) -> float:
    r"""Round a given value to Integer, under Power-of-2 restrction.

    ### Formula:

    ppq_round_to_power_of_2(x) = sign(x) * float(pow(2, round(log2(sign(x) * x))))
        where sign is +1 or -1
    """
    if value == 0:
        return 0
    sign = 1 if value >= 0 else -1
    assert isinstance(value, float) or isinstance(
        value, int
    ), "power-of-2 round only takes effect on float or int."
    return sign * float(pow(2, ppq_numerical_round(log2(sign * value), policy=policy)))



from typing import Iterable, Optional

from mppq.defs import empty_ppq_cache
from mppq.executor.base import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.quantize import QuantableOperation
from mppq.quantization.optim.base import OPTIM_ALGORITHMS, QuantizationOptimizationPass
from mppq.utils.qfunction import ppq_fake_quant


@OPTIM_ALGORITHMS.register()
class ParameterBakingPass(QuantizationOptimizationPass):
    r"""将计算图中所有已量化算子的参数进行烘焙。
    烘焙后的算子将省去伪量化的计算，可以加速约20%。

    Note:

        烘焙后的计算图不可逆
    """

    def __init__(self) -> None:
        super().__init__()
        self._quantize_function = ppq_fake_quant

    @empty_ppq_cache
    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ):
        for _, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue
            operation.baking_parameters(self._quantize_function)



from typing import Optional, Tuple

import torch
from torch.ao.quantization.observer import HistogramObserver

from mppq.common import (
    OBSERVER_KL_COMPUTING_DEVICE,
    OBSERVER_KL_HIST_BINS,
    OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE,
    OBSERVER_MIN_SCALE,
    OBSERVER_MIN_SCALE_MANUL_OVERRIDE,
)
from mppq.defs import ppq_quant_param_computing_function
from mppq.ffi import CUDA, ENABLE_CUDA_KERNEL
from mppq.ir.base.opdef import Variable
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TensorQuantizationConfig,
)
from mppq.quantization.measure.statistic import torch_KL_divergence
from mppq.quantization.observer.base import OBSERVER_TABLE, BaseTensorObserver
from mppq.quantization.observer.min_max import TorchMinMaxObserver
from mppq.utils.round import ppq_round_to_power_of_2


@OBSERVER_TABLE.register("kl")
class TorchHistObserver(TorchMinMaxObserver):
    r"""TorchHistObserver collects histogram of given tensor.

    It is designed for per-tensor quantization or activation quantization.
    """

    def __init__(
        self,
        watch_on: Variable,
        quant_cfg: TensorQuantizationConfig,
        hist_bins: int = OBSERVER_KL_HIST_BINS,
    ):
        self._phase = "Detecting Minmax"
        self._hist: Optional[torch.Tensor] = None
        self._hist_scale: float = 1
        self._min: float = 0
        self._max: float = 0
        if OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE in quant_cfg.detail:
            hist_bins = quant_cfg.detail[OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE]
        self._hist_bins: int = hist_bins
        super().__init__(watch_on, quant_cfg)

    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert (
            value.numel() > 0
        ), f"You are observing an empty tensor({self._watch_on.name})."

        if self._phase == "Detecting Minmax":
            return super().observe(value)  # collect min, max

        elif self._phase == "Collating Hist":
            if self._hist is None:
                self._hist = torch.zeros(
                    size=(self._hist_bins,), dtype=torch.int32, device=value.device
                )

            if self._quant_cfg.policy.has_property(QuantizationProperty.ASYMMETRIC):
                # ASYMMETRICAL Hist
                if ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL and value.is_cuda:
                    CUDA.Histogram_Asymmetric_T(
                        self._min, self._max, tensor=value, histogram=self._hist
                    )
                else:
                    hist = torch.histc(
                        value, self._hist_bins, min=self._min, max=self._max
                    )
                    self._hist += hist.int()

            elif self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRIC):
                # SYMMETRICAL Hist
                if ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL and value.is_cuda:
                    CUDA.Histogram_T(
                        tensor=value, histogram=self._hist, scale=self._hist_scale
                    )
                else:
                    hist = torch.histc(
                        torch.abs(value),
                        self._hist_bins,
                        min=0,
                        max=self._hist_scale * self._hist_bins,
                    )
                    self._hist += hist.int()

            else:
                raise TypeError(
                    "Quantization Property is invalid, "
                    "expect either ASYMMETRICAL or SYMMETRICAL config here."
                )

    @ppq_quant_param_computing_function
    def hist_to_scale_offset(
        self,
        histogram: torch.Tensor,
        hist_bins: int,
        hist_scale: float,
        config: TensorQuantizationConfig,
        computing_device: str = OBSERVER_KL_COMPUTING_DEVICE,
        scale_threshold: float = OBSERVER_MIN_SCALE,
    ) -> Tuple[float, int]:
        """
        PPQ core quant parameter computing method - Histogram to scale & offset

        With a pre-defined histogram,
        this function will automatically search best clip value
        to minimize KL divergence between quantized result and fp32 input.

        only work for per-tensor symmetrical quantization policy for now.
        see also

        https://on-demand.gputechconf.com/gtc/2017/presentation/s7310-8-bit-inference-with-tensorrt.pdf

        Args:
            histogram (torch.Tensor): histogram records activation's statistics.
            hist_bins (int): how many bins are included in histogram
                (also known as histogram length)
            hist_scale (float): histogram step size. it can be solved by
                histogram.max_val / histogram.bins
            config (TensorQuantizationConfig): quantization config.
            computing_device (str, optional): computing device. Defaults to 'cpu'.

        Raises:
            ValueError: given quantization config is invalid.

        Returns:
            Tuple[float, int]: scale(fp32) and offset(int).
        """
        if config.policy.has_property(QuantizationProperty.ASYMMETRIC):
            raise PermissionError(
                "KL observer is not designed for ASYMMETRICAL quantization"
            )

        if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
            scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]

        # move histogram to cpu, speedup computation.
        histogram = histogram.to(computing_device).float()

        # compute symmtrical kl-divergence.
        # Here is a simple example: reference distribution P consisting of 8 bins,
        # we want to quantize into 2 bins:
        # P = [ 1, 0, 2, 3, 5, 3, 1, 7]
        # we merge into 2 bins (8 / 2 = 4 consecutive bins are merged into one bin)
        # [1 + 0 + 2 + 3 , 5 + 3 + 1 + 7] = [6, 16]
        # then proportionally expand back to 8 bins, we preserve empty bins from the
        # original distribution P:
        # Q = [ 6/3, 0, 6/3, 6/3, 16/4, 16/4, 16/4, 16/4] = [ 2, 0, 2, 2, 4, 4, 4, 4]
        # now we should normalize both distributions, after that we can compute
        # KL_divergence
        # P /= sum(P) Q /= sum(Q)
        # result = KL_divergence(P, Q)
        # see also
        # https://github.com/NVIDIA/TensorRT/blob/3835424af081db4dc8cfa3ff3c9f4a8b89844421/tools/pytorch-quantization/pytorch_quantization/calib/histogram.py#L147

        losses, quant_bins = [], 2 ** (config.num_of_bits - 1)

        # following code is curcial, do not remove
        histogram[: int(hist_bins * 0.002)] = 0
        histogram[int(hist_bins * 0.002)] = 1

        hist_sum = torch.sum(histogram)
        for bin_range in range(quant_bins, hist_bins + quant_bins - 1, quant_bins):
            p_hist = torch.zeros(
                size=(bin_range,), dtype=torch.float, device=computing_device
            )
            p_hist[:bin_range].copy_(histogram[:bin_range])
            p_hist[bin_range - 1] += torch.sum(histogram[bin_range:])
            p_hist = p_hist / hist_sum

            expand_ratio = int(bin_range / quant_bins)
            q_hist = histogram[:bin_range].clone()
            q_hist = q_hist.reshape((quant_bins, expand_ratio))
            positive_map = q_hist > 0
            positive_cnt = positive_map.sum(dim=1, keepdim=True)
            positive_cnt[positive_cnt == 0] = 1
            q_hist = torch.div(q_hist.sum(dim=1, keepdim=True), positive_cnt)
            q_hist = q_hist.repeat([1, expand_ratio])
            q_hist = q_hist * positive_map
            q_hist = q_hist / torch.sum(q_hist)
            q_hist = q_hist.flatten()

            losses.append(
                {"kl": torch_KL_divergence(p_hist, q_hist), "bin_range": bin_range}
            )

        best_bin_range = sorted(losses, key=lambda x: x["kl"])[0]["bin_range"]
        scale, offset = (best_bin_range / self._hist_bins) * hist_scale * (
            self._hist_bins / quant_bins
        ), 0

        if scale < scale_threshold:
            self._logger.warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)

        if config.policy.has_property(QuantizationProperty.POWER_OF_2):
            scale = ppq_round_to_power_of_2(scale, policy=RoundingPolicy.ROUND_HALF_UP)
        return scale, offset

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if not self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            raise ValueError(
                "Hist observer can only apply with per-tensor quantization config."
            )

        if self._phase == "Detecting Minmax":
            min_val = torch.min(torch.cat(self._min_val_collector, dim=0)).cpu().item()
            max_val = torch.max(torch.cat(self._max_val_collector, dim=0)).cpu().item()
            if self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRIC):
                hist_range = float(max(abs(max_val), abs(min_val)))
            else:
                hist_range = max_val - min_val
            self._min = min_val
            self._max = max_val
            self._hist_scale = hist_range / self._hist_bins
            self._phase = "Collating Hist"
        elif self._phase == "Collating Hist":
            assert self._hist is not None
            scale, offset = self.hist_to_scale_offset(
                histogram=self._hist,
                hist_bins=self._hist_bins,
                hist_scale=self._hist_scale,
                config=self._quant_cfg,
            )
            device = self._hist.device
            self._quant_cfg.scale = torch.tensor(
                [scale], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.offset = torch.tensor(
                [offset], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED


@OBSERVER_TABLE.register(name="hist")
class TorchHistogramObserver(BaseTensorObserver):
    """Porting torch original hist observer."""

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        if OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE in quant_cfg.detail:
            hist_bins = quant_cfg.detail[OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE]
        else:
            hist_bins = OBSERVER_KL_HIST_BINS
        if quant_cfg.policy.has_property(QuantizationProperty.ASYMMETRIC):
            dtype = torch.quint8
            qscheme = torch.per_tensor_affine
        else:
            dtype = torch.qint8
            qscheme = torch.per_tensor_symmetric
        self._ob = HistogramObserver(hist_bins, dtype=dtype, qscheme=qscheme)

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert (
            value.numel() > 0
        ), f"You are observing an empty tensor({self._watch_on.name})."

        if self._ob.histogram.device != value.device:
            self._ob.to(value.device)
        self._ob(value)

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if not self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            raise ValueError(
                "Hist observer can only apply with per-tensor quantization config."
            )

        scales, zero_points = self._ob.calculate_qparams()
        device = self._ob.histogram.device
        self._quant_cfg.scale = scales.to(device=device).squeeze(0)
        self._quant_cfg.offset = zero_points.to(device=device).squeeze(0).float()
        self._quant_cfg.state = QuantizationStates.ACTIVATED



from typing import Tuple

import torch

from mppq.common import (
    OBSERVER_MIN_SCALE,
    OBSERVER_MIN_SCALE_MANUL_OVERRIDE,
    OBSERVER_MSE_COMPUTE_INTERVAL,
    OBSERVER_MSE_HIST_BINS,
)
from mppq.defs import ppq_quant_param_computing_function
from mppq.ffi import CUDA, ENABLE_CUDA_KERNEL
from mppq.ir.base.opdef import Variable
from mppq.quant import QuantizationProperty, TensorQuantizationConfig
from mppq.quantization.observer.base import OBSERVER_TABLE
from mppq.quantization.observer.hist import TorchHistObserver
from mppq.utils.qfunction.linear import minmax_to_scale_offset


@OBSERVER_TABLE.register("mse")
class TorchMSEObserver(TorchHistObserver):
    """Histogram accelerated MSE Observer, inspired by LightGBM This observer
    will collect data in histogram firstly, all mse computing will directly use
    histogram rather than data itself.

    Time complexity::

        O(Iteration * Num_of_Batch * Length(Data)) -> O(Iteration * Length(histogram))

    Space complexity::

        O(Num_of_Batch * Length(Data)) -> O(Iteration * Length(histogram))

    Args:
        TorchHistObserver ([type]): [description]
    """

    def __init__(
        self,
        watch_on: Variable,
        quant_cfg: TensorQuantizationConfig,
        bins: int = OBSERVER_MSE_HIST_BINS,
    ):
        super().__init__(watch_on, quant_cfg)
        self._hist_bins = bins

    def compute_mse_loss(self, histogram: list, start: int, step: int, end: int):
        if ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL:
            return CUDA.compute_mse_loss(
                histogram=histogram, start=start, step=step, end=end
            )
        else:
            # 如果你觉得 mse 太慢，想办法加速这段代码就可以了
            # 求解 mse 时，我们假设每一个 bin 里面的数据都是均匀分布的
            # 我们需要给一个直方图，并用 start, end, step 给出量化表示的范围
            # losses = [0 for _ in histogram]  debug
            num_of_elements = sum(histogram)
            loss = 0
            for idx, bin in enumerate(histogram):
                if idx < start:
                    # 如果所选的 bin 已经超出了起点，那从 bin 的中心到起点的距离即
                    # ((idx 到 起点的距离) + 0.5)
                    # 注意 hist 统计时是直接取 floor 的，因此会在这里额外 - 1
                    error = (start - idx - 1) + 0.5
                elif idx > end:
                    # 注意 hist 统计时是直接取 floor 的
                    error = (idx - end) + 0.5
                else:
                    # 分别计算左右两侧的 err
                    l_idx = (idx - start) % step
                    r_idx = step - l_idx - 1
                    if l_idx == r_idx:
                        error = l_idx + 0.25
                    else:
                        l_err = l_idx + 0.5
                        r_err = r_idx + 0.5
                        error = min(l_err, r_err)
                loss += (bin * error * error) / num_of_elements
                # losses[idx] = bin * error * error
            return loss

    @ppq_quant_param_computing_function
    def hist_to_scale_offset(  # type: ignore
        self,
        histogram: torch.Tensor,
        hist_bins: int,
        hist_scale: float,
        config: TensorQuantizationConfig,
        scale_threshold: float = OBSERVER_MIN_SCALE,
    ) -> Tuple[float, int]:
        if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
            scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]
        hist = histogram.tolist()
        num_of_quant_levels = (
            self._quant_cfg.quant_max - self._quant_cfg.quant_min
        ) + 1

        losses = []
        if config.policy.has_property(
            QuantizationProperty.ASYMMETRIC
        ) and config.policy.has_property(QuantizationProperty.PER_TENSOR):
            # at least we can have a min-max result
            step = hist_bins // num_of_quant_levels + 1
            loss = self.compute_mse_loss(
                histogram=hist, start=0, step=step, end=num_of_quant_levels * step
            )
            losses.append({"mse": loss, "start": 0, "end": num_of_quant_levels * step})

            for start in range(0, hist_bins, OBSERVER_MSE_COMPUTE_INTERVAL):
                if (start * hist_scale) + self._min > 0:
                    break  # start can not > 0

                for step in range(1, hist_bins // num_of_quant_levels + 1):
                    end = start + num_of_quant_levels * step
                    if end > (hist_bins + num_of_quant_levels):
                        break
                    loss = self.compute_mse_loss(
                        histogram=hist, start=start, step=step, end=end
                    )
                    losses.append({"mse": loss, "start": start, "end": end})

            best_policy = sorted(losses, key=lambda x: x["mse"])[0]
            best_start = best_policy["start"]
            best_end = best_policy["end"]

            # translate start & end to scale & offset.
            range_min, range_max = (best_start * hist_scale) + self._min, (
                best_end * hist_scale
            ) + self._min
            scale, offset = minmax_to_scale_offset(
                range_min, range_max, config, scale_threshold
            )
            return scale, int(offset)

        elif config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            raise PermissionError(
                "Torch Mse observer do not support PER_CHANNEL policy now, please wait."
            )

        elif config.policy.has_property(
            QuantizationProperty.SYMMETRIC
        ) and config.policy.has_property(QuantizationProperty.PER_TENSOR):
            # at least we can have a min-max result
            step = hist_bins // num_of_quant_levels + 1
            loss = self.compute_mse_loss(
                histogram=hist, start=0, step=step, end=num_of_quant_levels * step
            )
            losses.append({"mse": loss, "end": num_of_quant_levels * step})

            for step in range(1, hist_bins // num_of_quant_levels + 1):
                end = num_of_quant_levels * step
                if end > (hist_bins + num_of_quant_levels):
                    break
                loss = self.compute_mse_loss(
                    histogram=hist, start=0, step=step, end=end
                )
                losses.append({"mse": loss, "end": end})

            best_policy = sorted(losses, key=lambda x: x["mse"])[0]
            best_end = best_policy["end"]

            # translate start & end to scale & offset.
            range_min, range_max = -(best_end * hist_scale), (best_end * hist_scale)
            scale, offset = minmax_to_scale_offset(
                range_min, range_max, config, scale_threshold
            )
            return scale, int(offset)

        raise RuntimeError("Oops, there might be some mistakes.")



from typing import List

import torch

from mppq.common import OBSERVER_PERCENTILE, OBSERVER_PERCENTILE_MANUL_OVERRIDE
from mppq.ffi import CUDA, ENABLE_CUDA_KERNEL
from mppq.ir.base.opdef import Variable
from mppq.logger import debug
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from mppq.quantization.observer.base import OBSERVER_TABLE, BaseTensorObserver
from mppq.utils.qfunction.linear import minmax_to_scale_offset


@OBSERVER_TABLE.register("percentile")
class TorchPercentileObserver(BaseTensorObserver):
    """TorchPercentileObserver collects percentile data of given tensor.

    It is designed for activation quantization.
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        if OBSERVER_PERCENTILE_MANUL_OVERRIDE not in quant_cfg.detail:
            self._percentile = OBSERVER_PERCENTILE
        else:
            self._percentile = quant_cfg.detail[OBSERVER_PERCENTILE_MANUL_OVERRIDE]
        self._percentile_collector: List[torch.Tensor] = []
        self._percentile_maxs: List[torch.Tensor] = []
        self._percentile_mins: List[torch.Tensor] = []

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert value is not None, (
            "You are observing an Empty Tensor. "
            "(This Error is usually due to you have a wrong Quantizer configuration.)"
        )
        assert (
            value.numel() > 0
        ), f"You are observing an empty tensor({self._watch_on.name})."
        assert isinstance(
            value, torch.Tensor
        ), "TorchMinMaxObserver can only deal with torch Tensor values"

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            if not ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL or (not value.is_cuda):
                numel = value.numel()

                min_idx, max_idx = int(numel * (1 - self._percentile)), int(
                    numel * (self._percentile)
                )
                # torch.kthvalue needs index from 1 to numel ...
                min_idx = max(0, min_idx) + 1
                max_idx = min(max_idx, numel - 1) + 1
                _min = torch.kthvalue(value.flatten(), k=min_idx, dim=0)[0].view(1, -1)
                _max = torch.kthvalue(value.flatten(), k=max_idx, dim=0)[0].view(1, -1)
                self._percentile_collector.append(torch.cat([_max, _min], dim=-1))
            else:
                self._percentile_collector.append(
                    CUDA.Quantile(value, self._percentile).view(1, -1)
                )
        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            channel_axis = self._quant_cfg.channel_axis
            channelwise_view = value.transpose(dim0=0, dim1=channel_axis)
            channelwise_view = torch.flatten(channelwise_view, start_dim=1)
            self._percentile_mins.append(
                -torch.quantile(
                    -channelwise_view, q=self._percentile, dim=1, keepdim=True
                )[0]
            )
            self._percentile_maxs.append(
                torch.quantile(
                    channelwise_view, q=self._percentile, dim=1, keepdim=True
                )[0]
            )
        else:
            raise TypeError(
                "Percentile Observer only work with per-tensor or per-channel "
                "quantize policy."
            )

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            if len(self._percentile_collector) == 0:
                raise ValueError(
                    "Can not render quantization config yet, "
                    "Observer data collator is empty. "
                    "Invoke observe() function before render config."
                )
            device = self._percentile_collector[-1].device
            percentile_collector = (
                torch.cat(self._percentile_collector, dim=0).float().mean(dim=0).cpu()
            )
            min_val = percentile_collector[1].item()
            max_val = percentile_collector[0].item()
            scale, offset = minmax_to_scale_offset(
                min_val=min_val, max_val=max_val, config=self._quant_cfg
            )
            debug(
                f"{self._watch_on.name} observes min_val={min_val}, max_val={max_val}, "
                f"scale={scale}, zero point={offset}"
            )

            self._quant_cfg.scale = torch.tensor(
                [scale], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.offset = torch.tensor(
                [offset], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            if len(self._percentile_maxs) == 0:
                raise ValueError(
                    "Can not render quantization config yet, "
                    "Observer data collator is empty. "
                    "Invoke observe() function before render config."
                )
            device = self._percentile_maxs[-1].device

            min_vals = torch.mean(
                torch.cat(self._percentile_mins, dim=-1), dim=-1, keepdim=False
            )
            max_vals = torch.mean(
                torch.cat(self._percentile_maxs, dim=-1), dim=-1, keepdim=False
            )

            min_vals = min_vals.cpu()
            max_vals = max_vals.cpu()

            assert len(min_vals) == len(
                max_vals
            ), "Min values and max values should at same length."
            scales, offsets = [], []
            for min_val, max_val in zip(min_vals, max_vals):
                scale, offset = minmax_to_scale_offset(
                    min_val=float(min_val),
                    max_val=float(max_val),
                    config=self._quant_cfg,
                )
                scales.append(scale)
                offsets.append(offset)

            self._quant_cfg.scale = torch.tensor(
                scales, dtype=torch.float32, device=device
            )
            self._quant_cfg.offset = torch.tensor(
                offsets, dtype=torch.int32, device=device
            )
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        else:
            raise TypeError(
                "Percentile Observer only work with per-tensor or per-channel "
                "quantize policy."
            )



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

PPQ frontend parser.

Currently support parser:
- onnx
- onnxruntime (onnx with Microsoft domain operators)
- openvino
"""

from importlib import import_module
from pathlib import Path

from mppq.frontend.base import EXPORTER, PARSER

# auto scan directory and load
for folder in Path(__file__).parent.iterdir():
    if folder.is_dir() and not folder.name.startswith("_"):
        import_module(f".{folder.name}", package=__name__)


__all__ = ["EXPORTER", "PARSER"]



import torch
from torch.autograd import Function

from mppq.ffi import CUDA, ENABLE_CUDA_KERNEL
from mppq.quant import QuantizationProperty, RoundingPolicy, TensorQuantizationConfig


# pylint: disable=abstract-method, arguments-differ
class TensorwiseFloatingQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        exponet_bits: int,
        mantissa_bits: int,
        quant_min: float,
        quant_max: float,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:

        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implementation
            raise NotImplementedError("This Feature must run with PPQ Cuda Kernel.")
        else:
            # quantization function, pure cuda implementation
            quantized = CUDA.FloatingQuantize_T(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                exponent=exponet_bits,
                mantissa=mantissa_bits,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value,
            )
            return quantized

    @staticmethod
    def backward(ctx, *dy: torch.Tensor):
        return dy[0], None, None, None, None, None, None, None, None


class ChannelwiseFloatingQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        exponet_bits: int,
        mantissa_bits: int,
        quant_min: float,
        quant_max: float,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:

        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            raise NotImplementedError("This Feature must run with PPQ Cuda Kernel.")
        else:
            quantized = CUDA.FloatingQuantize_C(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                channel_axis=channel_axis,
                exponent=exponet_bits,
                mantissa=mantissa_bits,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value,
            )
            return quantized

    @staticmethod
    def backward(ctx, *dy: torch.Tensor):
        return dy[0], None, None, None, None, None, None, None, None, None


def floating_quant(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        qtensor = ChannelwiseFloatingQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.channel_axis,
            config.exponent_bits,
            config.mantissa_bits,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    else:
        qtensor = TensorwiseFloatingQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.exponent_bits,
            config.mantissa_bits,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    assert isinstance(qtensor, torch.Tensor)
    return qtensor



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Default quantizer for example.
"""

from mppq.common import COMPUTING_OP
from mppq.ir.base.graph import Operation
from mppq.quant import (
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPrecision,
    TensorQuantizationConfig,
)
from mppq.quantization.optim.base import (
    OPTIM_ALGORITHMS,
    QuantizationOptimizationPipeline,
)
from mppq.quantizer.base import QUANTIZER, BaseQuantizer


def create_default_quant_config(
    op: Operation,
    num_of_bits: int = 8,
    quant_min: int = 0,
    quant_max: int = 255,
    observer_algorithm: str = "percentile",
    policy: QuantizationPolicy = QuantizationPolicy(
        QuantizationProperty.PER_TENSOR
        | QuantizationProperty.LINEAR
        | QuantizationProperty.ASYMMETRIC
    ),
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
    exponent_bits: int = 0,
) -> OperationQuantizationConfig:
    r"""为你的算子创建一个默认量化信息

    对于一个 Onnx 算子而言，它总是会有几个输入和输出 Variable
    你需要为每一个相连的 Variable 初始化量化信息 TensorQuantConfig

    这个函数就是用来帮你初始化这些信息的。

    一个麻烦的问题是：

        对于很多 onnx 算子而言，他们的部分输入都是不需要量化的:

        如 Clip 算子的三个输入 value, min, max, 大部分框架不要求量化 min, max
        如 Reshape 算子的两个输入 value, shape, 其中 shape 不能够被量化

        PPQ 的算子接线器中记录了这些信息

        算子接线器中记录了所有标准 onnx 的默认量化策略
        该函数将使用预定义的算子量化策略初始化量化信息

    你可以在 Quantizer 中对默认量化策略进行进一步修改

    Create a default quantization configuration for given op.
    For each onnx op, there will be some input and output variables.

    You are required to create tensor quantization config for every
    input and output variables.

    This function is designed for creating a default quantization config for you.

        The created OQC(Op Quantization Config) is based on OpSocket.

        In fact, there are some rules or templates when creating the OQC:
        For Clip Op which has 3 input variable, namely value, min and max
            most framework does not require a quantization config for min and max.
        For Reshape Op which has 2 input variable, namely value and shape
            the input shape can never be quantized.

    Those rules are pre-defined within OpSocket, thus ppq will create
    OQC based on underlying OpSocket of your Op.

    After the default OQC got created, you can overwrite its state in quantizer.
    """
    assert isinstance(
        op, Operation
    ), f"Can only initialize OQC for PPQ.IR.Operation, however {type(op)} was given."
    assert isinstance(
        policy, QuantizationPolicy
    ), "Can not create quantization config - Quantization Policy Type Error."
    assert isinstance(
        rounding, RoundingPolicy
    ), "Can not create quantization config - Rounding Policy Type Error."

    socket = op.socket
    input_cfgs, output_cfgs = [], []
    for index in range(op.num_of_input):
        state = QuantizationStates.INITIAL
        # for those unexpected inputs and outputs
        # ppq just initialize them as normal variable.
        if index < len(socket.in_plat):
            target_plat = socket.in_plat[index]
            if target_plat == TargetPrecision.FP32:
                state = QuantizationStates.FP32
            if target_plat == TargetPrecision.SOI:
                state = QuantizationStates.FP32
        input_cfgs.append(
            TensorQuantizationConfig(
                policy=policy,
                rounding=rounding,
                num_of_bits=num_of_bits,
                scale=None,
                offset=None,
                exponent_bits=exponent_bits,
                quant_min=quant_min,
                quant_max=quant_max,
                observer_algorithm=observer_algorithm,
                state=state,
            )
        )

    for index in range(op.num_of_output):
        state = QuantizationStates.INITIAL
        # for those unexpected inputs and outputs
        # ppq just initialize them as normal variable.
        if index < len(socket.out_plat):
            target_plat = socket.out_plat[index]
            if target_plat == TargetPrecision.FP32:
                state = QuantizationStates.FP32
            if target_plat == TargetPrecision.SOI:
                state = QuantizationStates.FP32
        output_cfgs.append(
            TensorQuantizationConfig(
                policy=policy,
                rounding=rounding,
                num_of_bits=num_of_bits,
                scale=None,
                offset=None,
                exponent_bits=exponent_bits,
                quant_min=quant_min,
                quant_max=quant_max,
                observer_algorithm=observer_algorithm,
                state=state,
            )
        )

    return OperationQuantizationConfig(input_cfgs, output_cfgs)


@QUANTIZER.register("default")
class DefaultQuantizer(BaseQuantizer):
    r"""默认量化器，用于示例。"""

    @property
    def quant_operation_types(self):
        return set(COMPUTING_OP)

    def init_quantize_config(self, operation):
        assert operation.type in self.quant_operation_types

        base_qc = create_default_quant_config(operation)
        weight_policy = QuantizationPolicy(
            QuantizationProperty.PER_CHANNEL
            | QuantizationProperty.LINEAR
            | QuantizationProperty.SYMMETRIC
        )
        weight_min, weight_max = -128, 127
        if operation.type in {"Conv", "ConvTranspose", "Gemm", "MatMul"}:
            # base_qc.output_quantization_config[0].state = QuantizationStates.FP32
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel
            # quant-config.
            assert (
                operation.num_of_input > 0
            ), f"Seems you got a {operation.type} layer with no parameters."

            # first parameter must exist, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {"Conv", "ConvTranspose"}:
                if operation.inputs[1].is_parameter:
                    conv_weight_config = base_qc.input_quantization_config[1]
                    conv_weight_config.quant_min = weight_min
                    conv_weight_config.quant_max = weight_max
                    conv_weight_config.policy = weight_policy
                    conv_weight_config.channel_axis = 1
                    if operation.type == "ConvTranspose":
                        conv_weight_config.channel_axis = 0
                    conv_weight_config.observer_algorithm = "minmax"
            # first parameter must exist, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {"Gemm", "MatMul"}:
                if operation.inputs[1].is_parameter:
                    gemm_weight_config = base_qc.input_quantization_config[1]
                    gemm_weight_config.quant_min = weight_min
                    gemm_weight_config.quant_max = weight_max
                    gemm_weight_config.policy = weight_policy
                    gemm_weight_config.channel_axis = 1
                    if operation.type == "Gemm" and operation.attributes.get(
                        "transB", None
                    ):
                        gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = "minmax"
            if operation.num_of_input > 2:
                bias_config = base_qc.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32
        elif operation.type == "LayerNormalization":
            # Layernorm - gamma and beta need to be FP32
            for qc in base_qc.input_quantization_config[1:]:
                qc.state = QuantizationStates.FP32
        return base_qc

    @property
    def default_prequant_pipeline(self) -> QuantizationOptimizationPipeline:
        return QuantizationOptimizationPipeline(
            [
                OPTIM_ALGORITHMS["ActivationEqualizationPass"](),
                OPTIM_ALGORITHMS["LayerwiseEqualizationPass"](),
                OPTIM_ALGORITHMS["ChannelwiseSplitPass"](),
            ]
        )

    @property
    def default_quant_pipeline(self) -> QuantizationOptimizationPipeline:
        return QuantizationOptimizationPipeline(
            [
                OPTIM_ALGORITHMS["SSDEqualizationPass"](),
                OPTIM_ALGORITHMS["QuantizeFusionPass"](),
                OPTIM_ALGORITHMS["QuantizeSimplifyPass"](),
                OPTIM_ALGORITHMS["ParameterQuantizePass"](),
                OPTIM_ALGORITHMS["IsotoneCalibrationPass"](),
                OPTIM_ALGORITHMS["RuntimeCalibrationPass"](),
                OPTIM_ALGORITHMS["QuantAlignmentPass"](),
                OPTIM_ALGORITHMS["PassiveParameterQuantizePass"](),
                OPTIM_ALGORITHMS["BiasCorrectionPass"](),
                OPTIM_ALGORITHMS["LearnedStepSizePass"](),
                OPTIM_ALGORITHMS["AdaroundPass"](),
                OPTIM_ALGORITHMS["RoundTuningPass"](),
                OPTIM_ALGORITHMS["PassiveParameterQuantizePass"](),
                OPTIM_ALGORITHMS["ParameterBakingPass"](),
            ]
        )



"""PPQ Core Foreign Function Interface PPQ 核心编程语言接口.

You are not allowed to modify this 请勿修改此文件
"""

# pylint: disable=invalid-name

import json
import os
import shutil
import subprocess as sp
from contextlib import contextmanager
from pathlib import Path
from typing import List, Optional

import pooch
import torch
from torch.cuda import synchronize
from torch.utils.cpp_extension import load

from mppq.defs import SingletonMeta
from mppq.logger import warning

_URLS = [
    (
        "https://github.com/microsoft/vswhere/releases/download/3.1.7/vswhere.exe",
        "c54f3b7c9164ea9a0db8641e81ecdda80c2664ef5a47c4191406f848cc07c662",
    ),
]


class MSVCLocator:
    """Download vswhere to locate the installed Microsoft Visual Studio instances.
    And set CXX environment variable to compiler path.

    Set CXX to None if no Visual Studio is installed.
    Raise FileNotFoundError if Visual Studio is installed without C++ components.
    """

    def __init__(self, arch: str = "x64"):
        self._old_cxx: Optional[str] = None
        if (vswhere := shutil.which("vswhere")) is None:
            for url, sha256 in _URLS:
                vswhere = pooch.retrieve(url, known_hash=sha256, progressbar=True)
        assert vswhere is not None
        self.vswhere = Path(vswhere)
        assert self.vswhere.exists() and self.vswhere.is_file()
        if arch.lower() in ("amd64", "x64"):
            self.x86 = False
        elif arch.lower() in ("x86", "i386"):
            self.x86 = True
        else:
            raise NotImplementedError(f"Unrecoganized arch: {arch}")

    def __enter__(self) -> Path:
        stdout = sp.check_output(
            ["", "-latest", "-utf8", "-format", "json"], executable=self.vswhere
        )
        results = json.loads(stdout)
        for cl in filter(
            lambda i: i is not None, (self._find_cl(search) for search in results)
        ):
            assert cl is not None
            self._old_cxx = os.environ.get("CXX")
            os.environ["CXX"] = str(cl.resolve())
            # expand path
            os.environ["PATH"] += f";{cl.parent.resolve()}"
            print(f"set CXX={cl}")
            break
        return Path(os.environ["CXX"])

    def __exit__(self, *args):
        if self._old_cxx is None and "CXX" in os.environ:
            os.environ.pop("CXX")
        elif self._old_cxx:
            os.environ["CXX"] = str(self._old_cxx)

    def _find_cl(self, search: dict) -> Optional[Path]:
        msvc_path = Path(search["installationPath"]) / "VC/Tools/MSVC"
        if not msvc_path.exists():
            raise FileNotFoundError(
                f"Visual Studio is installed without C++ components: {msvc_path}"
            )
        for instance_dir in filter(lambda p: p.is_dir(), msvc_path.glob("*")):
            if self.x86:
                cl = instance_dir / "bin/Hostx86/x86/cl.exe"
            else:
                cl = instance_dir / "bin/Hostx64/x64/cl.exe"
            if cl.exists() and cl.is_file():
                return cl
        return None


class CompilerHelper(metaclass=SingletonMeta):
    """PPQ-Torch Compile Wrapper."""

    def __init__(self) -> None:
        self.__cuda_extention__ = None

    def compile(self):
        """Compile CUDA kernels."""

        warning("Compiling Kernels... Please wait (It will take a few minutes).")
        build_directory = Path(__file__).parent / "csrc/build/"
        lock_file = build_directory / "lock"

        if not build_directory.exists():
            build_directory.mkdir(parents=True)
        if os.path.exists(lock_file):
            try:
                os.remove(lock_file)
            except Exception as e:
                raise PermissionError(
                    f"Can not delete lock file at {lock_file}, delete it first!"
                ) from e

        self.__cuda_extention__ = load(
            name="PPQ_Cuda_Impls",
            sources=[
                str(Path(__file__).parent / "csrc/export.cc"),
                str(Path(__file__).parent / "csrc/cuda/linear.cu"),
                str(Path(__file__).parent / "csrc/cuda/sort.cu"),
                str(Path(__file__).parent / "csrc/cuda/train.cu"),
                str(Path(__file__).parent / "csrc/cuda/floating.cu"),
                str(Path(__file__).parent / "csrc/cpu/hist_mse.cc"),
            ],
            build_directory=build_directory,
            with_cuda=True,
            extra_cflags=["-O3"],
        )

    @property
    def extension(self):
        """Return the loaded CUDA library."""
        if self.__cuda_extention__ is None:
            raise RuntimeError(
                "Cuda Extension has not been compiled, "
                "invoke ppq.core.ffi.ComplieHelper.compile() First."
            )
        return self.__cuda_extention__


CUDA_COMPLIER = CompilerHelper()


@contextmanager
def dummy_locator():
    """Linux compiler locator, do nothing for now"""
    yield


class ENABLE_CUDA_KERNEL:
    """Auto config compiler path before entering compiling CUDA context"""

    USING_CUDA_KERNEL = False

    def __init__(self) -> None:
        self._state = True
        if os.name == "nt":
            with MSVCLocator():
                self._compile()
        else:
            with dummy_locator():
                self._compile()

    def _compile(self):
        if CUDA_COMPLIER.__cuda_extention__ is None:
            CUDA_COMPLIER.compile()
        self._state = False

    def __enter__(self):
        self._state = ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL
        ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL = True

    def __exit__(self, *args):
        ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL = self._state


# pylint: disable=missing-function-docstring, invalid-name
# helper class for calling cuda methods.
class CUDA:
    """CUDA is a helper class for invoking highly-effcient custimized cuda
    kernel. PPQ developer team has implemented a series of quantization related
    cuda kernel, They are 5-100x faster than torch kernels, with less gpu
    memory cost.

    You can easily extend your cuda kernel via this class:
        Firstly, implement your kernel within ppq/csrc/cuda, write your own .cu file
        and .h file.

        Secondly, add your functions to ppq/csrc/cuda/export.cc, add them to export
        table.

        Finally, add a interface with this python class(ppq.core.ffi.CUDA), following
        the signature as same as others.

    PPQ CUDA EXTENSION 命名规则:
        我们使用函数名+后缀名的形式命名 CUDA Extension 函数:

        后缀名 _T 表示 Tensorwise 函数
        后缀名 _C 表示 Channelwise 函数
        后缀名 _B 表示 导函数

    例如函数 LinearQuantize_T_B 表示线性量化函数的 Tensorwise 版本，并且是导函数。
    """

    @staticmethod
    def LinearQuantize_T(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        # if scale is too small, quantization might cause fp32 underflow.
        # if scale < 1e-7: raise ValueError('scale is too small.')
        return CUDA_COMPLIER.extension.QuantizeTensor_LT(
            tensor, scales, offsets, minimum, maximum, rounding
        )

    @staticmethod
    def LinearQuantize_C(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        return CUDA_COMPLIER.extension.QuantizeTensor_LC(
            tensor, scales, offsets, minimum, maximum, channel_axis, rounding
        )

    @staticmethod
    def LinearQuantize_T_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        minimum: int,
        maximum: int,
        rounding: int,
    ) -> List[torch.Tensor]:
        return CUDA_COMPLIER.extension.QuantizeTensor_LT_B(
            tensor, scales, offsets, dy, minimum, maximum, rounding
        )

    @staticmethod
    def LinearQuantize_C_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        minimum: int,
        maximum: int,
        channel_axis: int,
        rounding: int,
    ) -> List[torch.Tensor]:
        return CUDA_COMPLIER.extension.QuantizeTensor_LC_B(
            tensor, scales, offsets, dy, minimum, maximum, rounding, channel_axis
        )

    @staticmethod
    def Histogram_T(
        tensor: torch.Tensor,
        histogram: torch.Tensor,
        scale: float,
        clip_outliers: bool = True,
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.extension.Histogram_T(tensor, scale, clip_outliers, histogram)
        return histogram

    @staticmethod
    def Histogram_Asymmetric_T(
        min_value: float,
        max_value: float,
        tensor: torch.Tensor,
        histogram: torch.Tensor,
        clip_outliers: bool = True,
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.extension.Histogram_Asymmetric_T(
            min_value, max_value, tensor, clip_outliers, histogram
        )
        return histogram

    @staticmethod
    def Histogram_C(
        tensor: torch.Tensor,
        channel_axis: int,
        histogram: torch.Tensor,
        scale: float,
        clip_outliers: bool = True,
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.extension.Histogram_C(
            tensor, channel_axis, scale, clip_outliers, histogram
        )
        return histogram

    @staticmethod
    def Quantile(
        tensor: torch.Tensor,
        q: float,
    ) -> torch.Tensor:
        return CUDA_COMPLIER.extension.Quantile_T(tensor, q)

    @staticmethod
    def TensorClip_T(
        tensor: torch.Tensor,
        reference: torch.Tensor,
        limit: torch.Tensor,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        if not reference.is_contiguous():
            tensor = reference.contiguous()
        return CUDA_COMPLIER.extension.TensorClip_T(tensor, reference, limit)

    @staticmethod
    def TensorClip_C(
        tensor: torch.Tensor,
        reference: torch.Tensor,
        limit: torch.Tensor,
        channel_axis: int,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        if not reference.is_contiguous():
            tensor = reference.contiguous()
        return CUDA_COMPLIER.extension.TensorClip_C(
            tensor, reference, limit, channel_axis
        )

    @staticmethod
    def RoundingLoss_LT(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.RoundingLoss_LT(
            tensor, scales, offsets, minimum, maximum, rounding
        )

    @staticmethod
    def RoundingLoss_LT_B(
        tensor: torch.Tensor,
        dy: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.RoundingLoss_LT_B(
            tensor, dy, scales, offsets, minimum, maximum, rounding
        )

    @staticmethod
    def RoundingLoss_LC(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.RoundingLoss_LC(
            tensor, scales, offsets, minimum, maximum, channel_axis, rounding
        )

    @staticmethod
    def RoundingLoss_LC_B(
        tensor: torch.Tensor,
        dy: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.RoundingLoss_LC_B(
            tensor, dy, scales, offsets, minimum, maximum, channel_axis, rounding
        )

    @staticmethod
    def OrderPreservingObserve(
        tensor: torch.Tensor,
    ) -> torch.Tensor:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.RoundingLoss_LC_B(tensor)

    @staticmethod
    def compute_mse_loss(
        histogram: List[int], start: int, step: int, end: int
    ) -> float:
        return CUDA_COMPLIER.extension.compute_mse_loss(histogram, start, step, end)

    @staticmethod
    def FloatingQuantize_T(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        exponent: int = 4,
        mantissa: int = 3,
        minimum: float = -448,  # FP8 E4M3
        maximum: float = +448,
        rounding: int = 0,
    ) -> torch.Tensor:
        if exponent <= 0:
            raise ValueError("Floating Quantization requires exponent > 0")
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        # if scale is too small, quantization might cause fp32 underflow.
        # if scale < 1e-7: raise ValueError('scale is too small.')
        return CUDA_COMPLIER.extension.QuantizeTensor_FT(
            tensor, scales, offsets, exponent, mantissa, minimum, maximum, rounding
        )

    @staticmethod
    def FloatingQuantize_C(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        exponent: int = 4,
        mantissa: int = 3,
        minimum: float = -448,  # FP8 E4M3
        maximum: float = +448,
        rounding: int = 0,
    ) -> torch.Tensor:
        if exponent <= 0:
            raise ValueError("Floating Quantization requires exponent > 0")
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.QuantizeTensor_FC(
            tensor,
            scales,
            offsets,
            exponent,
            mantissa,
            minimum,
            maximum,
            channel_axis,
            rounding,
        )

    @staticmethod
    def FloatingQuantize_T_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        exponent: int = 4,
        mantissa: int = 3,
        minimum: float = -448,
        maximum: float = 448,
        rounding: int = 0,
    ) -> List[torch.Tensor]:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.QuantizeTensor_FT_B(
            tensor, scales, offsets, dy, exponent, mantissa, minimum, maximum, rounding
        )

    @staticmethod
    def FloatingQuantize_C_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        channel_axis: int,
        exponent: int = 4,
        mantissa: int = 3,
        minimum: float = -448,
        maximum: float = 448,
        rounding: int = 0,
    ) -> List[torch.Tensor]:
        if not tensor.is_contiguous():
            tensor = tensor.contiguous()
        return CUDA_COMPLIER.extension.QuantizeTensor_FC_B(
            tensor,
            scales,
            offsets,
            dy,
            exponent,
            mantissa,
            minimum,
            maximum,
            rounding,
            channel_axis,
        )

    @staticmethod
    def Sync():
        """Synchronize device."""
        synchronize()



from typing import List

import torch

from mppq.common import OBSERVER_ISOTONE_OBSERVER_AXIS
from mppq.ir.base.graph import Variable
from mppq.logger import warning
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from mppq.quantization.observer.base import OBSERVER_TABLE, BaseTensorObserver
from mppq.utils.qfunction.linear import minmax_to_scale_offset


@OBSERVER_TABLE.register("isotone")
class TorchIsotoneObserver(BaseTensorObserver):
    """For softmax or sigmoid activations, usually we just need
    argmax(softmax(x)) == argmax(softmax(quant(x)))

    Inspired by this Property, Isotone Observer is designed to provide an
    order-preserving calibration method, which cares only about argmax(x) [or argmin(x)]

    To keep argmax(x) == argmax(quant(x)), we only need to
    distinguish the largest element and the second largert element with quantization

    let L1 represents the largest element of x,
    while L2 represents the second largest.

    For symmetric quantization, We want

        1. L1 - L2 > scale
        2. round(L2 / scale) <= (quant_max - .5)
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        self._cache: List[torch.Tensor] = []
        if OBSERVER_ISOTONE_OBSERVER_AXIS not in quant_cfg.detail:
            warning(
                "Initializing TorchIsotoneObserver with implicit axis"
                " is not recommended."
            )
            self.axis = -1
        else:
            self.axis = quant_cfg.detail[OBSERVER_ISOTONE_OBSERVER_AXIS]
        self.s_candidates = None

    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert isinstance(
            value, torch.Tensor
        ), "IsotoneObserver can only deal with torch Tensor values"
        assert (
            value.numel() > 0
        ), f"You are observing an empty tensor({self._watch_on.name})."

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            # flatten value as [-1, num_of_elements in isotone axis]
            if value.ndim > 1:
                value = value.transpose(dim0=self.axis, dim1=-1)
                value = value.flatten(start_dim=0, end_dim=-2)
            value, _ = torch.topk(value, k=2, dim=self.axis, largest=True, sorted=True)
            if value.ndim <= 1:
                value = value.unsqueeze(0)
            self._cache.append(value)
        else:
            raise TypeError(
                "Isotone Observer only work with per-tensor quantize policy."
            )

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        device = self._cache[-1].device
        collected = torch.cat(self._cache, dim=0)
        collected = collected.cpu().numpy()
        s_candidates = []

        for l1, l2 in collected:
            if self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRIC):
                l1, l2 = abs(l1), abs(l2)

            scale_min = max(l2 / (self._quant_cfg.quant_max - 0.51), 0)
            scale_max = 2 * (l1 - max(l2, 0))
            if scale_max > scale_min and l1 > 0:
                s_candidates.append((scale_min, "min"))
                s_candidates.append((scale_max, "max"))

        if len(s_candidates) <= 0:
            max_val = collected[-1, 0]
            # fall back to min-max calibration
            scale, offset = minmax_to_scale_offset(
                min_val=0, max_val=max_val, config=self._quant_cfg
            )
            self._quant_cfg.scale = torch.tensor(
                [scale], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.offset = torch.tensor(
                [offset], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
            warning(
                "There is no way to make clear classification on "
                f"{self._watch_on.name} under int8 quantization."
            )
            return

        s_candidates = sorted(s_candidates)
        best_satisfied, best_scale, satisfied = 0, 0, 0
        for s_candidate, s_type in s_candidates:
            if s_type == "min":
                satisfied += 1
            if s_type == "max":
                satisfied -= 1

            if satisfied > best_satisfied:
                best_satisfied = satisfied
                best_scale = s_candidate

        self._quant_cfg.scale = torch.tensor(
            [best_scale], dtype=torch.float32, device=device
        ).squeeze(0)
        self._quant_cfg.offset = torch.tensor(
            [0], dtype=torch.float32, device=device
        ).squeeze(0)
        self._quant_cfg.state = QuantizationStates.ACTIVATED
        self.s_candidates = s_candidates



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Basic logging utilities for mppq.

Affected system environment variables:

- MPPQ_LOG_LEVEL: Set to specify initialized logging level.
- MPPQ_LOG_FILE: Set to stream log messages to a file.
"""

import logging
import os
from logging import Formatter, StreamHandler, addLevelName, getLevelName
from typing import Literal, Union, overload


def _default_level_from_env():
    match os.environ.get("MPPQ_LOG_LEVEL"):
        case "LOG_TRACE" | "0":
            return 1
        case "LOG_DEBUG" | "1":
            return logging.DEBUG
        case "LOG_INFO" | "2":
            return logging.INFO
        case "LOG_WARNING" | "3":
            return logging.WARNING
        case "LOG_ERROR" | "4":
            return logging.ERROR
    return logging.INFO


def trace(msg: str):
    """Logging trace message"""
    _LOG.log(1, msg)


def debug(msg: str):
    """Logging debug message"""
    _LOG.debug(msg)


def info(msg: str):
    """Logging informative message"""
    _LOG.info(msg)


def warning(msg: str):
    """Logging warning message"""
    _LOG.warning(msg)


def error(msg: str):
    """Logging error message"""
    _LOG.error(msg)


def fatal(msg: str):
    """Logging critical(fatal) message"""
    _LOG.critical(msg)


def set_level(level: str):
    """Filter log messages by different level."""
    _HDL.setLevel(level.upper())


def is_enabled_for(level: Union[int, str]) -> bool:
    """Whether current log level is active."""
    if isinstance(level, str):
        from logging import _nameToLevel  # pylint:disable=import-outside-toplevel

        return _LOG.isEnabledFor(_nameToLevel[level])
    return _LOG.isEnabledFor(level)


@overload
def get_level() -> str: ...  # noqa: E704


@overload
def get_level(use_string: Literal[True]) -> str: ...  # noqa: E704


@overload
def get_level(use_string: Literal[False]) -> int: ...  # noqa: E704


def get_level(use_string: bool = True) -> Union[int, str]:
    """Get current logging level."""
    if use_string:
        return getLevelName(_HDL.level)
    return _HDL.level


def drop():
    """Drop the logger handlers and flush pending events.
    Logging after `drop` will do nothing.
    """
    for hdl in list(_LOG.handlers):
        hdl.flush()
        hdl.close()
        _LOG.removeHandler(hdl)


def nest(nested: str):
    """Create a nested logger with nested name."""
    log = logging.getLogger(f"MPPQ.{nested}")
    log.setLevel(_LOG.level)
    return log


addLevelName(1, "TRACE")
_LOG = logging.getLogger("MPPQ")
_LOG.setLevel("TRACE")
_HDL = StreamHandler()
_HDL.setLevel(_default_level_from_env())
# use colorlog to colorize different level messages
try:
    import colorlog  # pylint:disable=import-outside-toplevel

    _HDL.setFormatter(
        colorlog.ColoredFormatter(
            "[%(asctime)s]%(log_color)s[%(name)s][%(levelname)s] %(message)s"
        )
    )
except ImportError:
    _HDL.setFormatter(Formatter("[%(asctime)s][%(name)s][%(levelname)s] %(message)s"))
_LOG.addHandler(_HDL)

if os.environ.get("MPPQ_LOG_FILE"):
    _FDL = logging.FileHandler(os.environ["MPPQ_LOG_FILE"], encoding="utf-8")
    _FDL.setFormatter(Formatter("[%(asctime)s][%(name)s][%(levelname)s] %(message)s"))
    _FDL.setLevel(_default_level_from_env())
    _LOG.addHandler(_FDL)



"""
Copyright Wenyi Tang 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

# pylint: disable=missing-function-docstring

import numpy as np
import pytest
import torch

from mppq.ir.base.opdef import Variable
from mppq.quant import (
    QuantizationPolicy,
    QuantizationProperty,
    TensorQuantizationConfig,
)
from mppq.quantization.observer.base import OBSERVER_TABLE


def per_tensor_affine():
    policy = QuantizationPolicy(
        QuantizationProperty.ASYMMETRIC
        | QuantizationProperty.LINEAR
        | QuantizationProperty.PER_TENSOR,
    )
    cfg = TensorQuantizationConfig(policy, quant_min=0, quant_max=255)
    return cfg


def per_tensor_symmetric():
    policy = QuantizationPolicy(
        QuantizationProperty.SYMMETRIC
        | QuantizationProperty.LINEAR
        | QuantizationProperty.PER_TENSOR,
    )
    cfg = TensorQuantizationConfig(policy, quant_min=-128, quant_max=127)
    return cfg


SIZE = 1000


def gaussian_dist(seed=42, mean=0, sigma=0.2):
    """Fixed gaussian distribution"""
    np.random.seed(seed)

    for _ in range(SIZE):
        yield np.random.normal(mean, sigma, size=[32])
    np.random.seed(None)


def gamma_dist(seed=42, shape=1, sigma=0.2):
    np.random.seed(seed)

    for _ in range(SIZE):
        yield np.random.gamma(shape, sigma, size=[32])
    np.random.seed(None)


def beta_dist(seed=42, alpha=0.2, beta=0.2):
    np.random.seed(seed)

    for _ in range(SIZE):
        yield np.random.beta(alpha, beta, size=[32])
    np.random.seed(None)


def chisquare_dist(seed=42, df=0.1):
    np.random.seed(seed)

    for _ in range(SIZE):
        yield np.random.chisquare(df, size=[32])
    np.random.seed(None)


@pytest.mark.parametrize("dist", [gaussian_dist, gamma_dist, beta_dist, chisquare_dist])
def test_all_observers_per_tensor_symmetric(dist):
    loss = {}
    for key in OBSERVER_TABLE:
        cls = OBSERVER_TABLE[key]
        try:
            observer = cls(Variable("a"), per_tensor_symmetric())
            for value in dist():
                observer.observe(torch.tensor(value).float())
            observer.render_quantization_config()
            # second phase, for observer without 2nd phase, do nothing
            for value in dist():
                observer.observe(torch.tensor(value).float())
            observer.render_quantization_config()
        except (TypeError, PermissionError):
            continue
        mse = 0
        for value in dist():
            scale = float(observer._quant_cfg.scale)
            zero_point = int(observer._quant_cfg.offset)
            qvalue = np.clip(
                np.round(value / scale) + zero_point,
                observer._quant_cfg.quant_min,
                observer._quant_cfg.quant_max,
            )
            qvalue = (qvalue - zero_point) * scale
            mse += np.sum((qvalue - value) ** 2)
        mse /= 10000
        loss[key] = mse
    print(f"\ntest_all_observers_per_tensor_symmetric {dist.__name__}")
    for name, mse in sorted(loss.items(), key=lambda pair: pair[1]):
        print(f"{name}: {mse}")
    print("")


@pytest.mark.parametrize("dist", [gaussian_dist, gamma_dist, beta_dist, chisquare_dist])
def test_possible_observers_per_tensor_affine(dist):
    loss = {}
    for key in OBSERVER_TABLE:
        cls = OBSERVER_TABLE[key]
        try:
            observer = cls(Variable("a"), per_tensor_affine())
            for value in dist():
                observer.observe(torch.tensor(value).float())
            observer.render_quantization_config()
            # second phase, for observer without 2nd phase, do nothing
            for value in dist():
                observer.observe(torch.tensor(value).float())
            observer.render_quantization_config()
        except (TypeError, PermissionError):
            continue
        mse = 0
        for value in dist():
            scale = float(observer._quant_cfg.scale)
            zero_point = int(observer._quant_cfg.offset)
            qvalue = np.clip(
                np.round(value / scale) + zero_point,
                observer._quant_cfg.quant_min,
                observer._quant_cfg.quant_max,
            )
            qvalue = (qvalue - zero_point) * scale
            mse += np.sum((qvalue - value) ** 2)
        mse /= 10000
        loss[key] = mse
    print(f"\ntest_possible_observers_per_tensor_affine {dist.__name__}")
    for name, mse in sorted(loss.items(), key=lambda pair: pair[1]):
        print(f"{name}: {mse}")
    print("")



from typing import Iterable, List, Literal, Optional, Set

import torch

from mppq.common import (
    ALIGNMENT_MANUL_OVERRIDE,
    PASSIVE_OPERATIONS,
    TYPES_FOR_ALIGNMENT,
)
from mppq.defs import empty_ppq_cache
from mppq.executor import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.quantize import QuantableOperation
from mppq.ir.search import SearchableGraph
from mppq.logger import warning
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from mppq.quantization.optim.base import OPTIM_ALGORITHMS, QuantizationOptimizationPass
from mppq.utils.qfunction.linear import minmax_to_scale_offset


@OPTIM_ALGORITHMS.register()
class QuantizeSimplifyPass(QuantizationOptimizationPass):
    r"""PPQ Quantize Simplify Pass(通用量化精简过程)

    PPQ use Tensor Quantization Configuration(A data structure defined in ppq.core) to
    control quantization. Each quantable op will have a list of TQC as its quantization
    config, which contains necessary quantization parameter(scale, offset), in order to
    quantize its input(s) and output(s).

    While TQC is a powerful tool for describing quantization, it introduces some
    undiserible features:

    For a subgraph like:

            Relu1 - Relu2

    PPQ will create at least 4 TQC here, namely the input TQC of Relu1 and Relu2, and
    the output TQC of Relu1 and Relu2. Problem here is the output TQC of Relu1 and the
    input TQC of Relu2 is actually duplicated, the output variable should not be
    quantized twice.

    This Simplify Pass will detect all the duplicated TQCs in your network, disable
    them and create a link with their dominating TQCs. Disabled TQC will have and
    inactive state(QuantizationState.OVERRLAPED), so PPQ executor will simply ignore
    them when executing.

    A duplicated TQC is an input TQC(A) whose binding variable has been quantized by
    another output TQC(B), and the input TQC(A) should have the same bit-width as the
    output TQC(B)

    ### Parameters:

    * No Parameter

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this
    optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True
        setting.fusion_setting.remove_useless_quantization = True

        # calling ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(self) -> None:
        super().__init__(name="PPQ Quantize Simplify Pass")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> None:
        for _, variable in graph.variables.items():
            assert isinstance(variable, Variable)
            source_op = variable.source_op

            if source_op is None:
                continue  # input variables in network, they do not have a source
            if not isinstance(source_op, QuantableOperation):
                continue
            source_config = source_op.config.output_quantization_config[
                source_op.outputs.index(variable)
            ]

            if source_config.state in {QuantizationStates.FP32}:
                continue  # if source config does not have a valid state, skip it.

            for downstream_op, dest_idx in zip(variable.dest_ops, variable.dest_idx):
                if downstream_op is None:
                    # output variables in network, they do not have a destination
                    continue
                if not isinstance(downstream_op, QuantableOperation):
                    continue

                input_config = downstream_op.config.input_quantization_config[dest_idx]
                if source_op.precision == downstream_op.precision:
                    if (
                        input_config.state == QuantizationStates.INITIAL
                        and input_config.is_same_scheme(source_config)
                    ):
                        input_config.dominated_by = source_config


@OPTIM_ALGORITHMS.register()
class QuantizeFusionPass(QuantizationOptimizationPass):
    r"""PPQ Quantize Fusion Pass(通用量化图融合过程)

    Operation fusion (or kernel/layer fusion) is key optimization in many SOTA
    execution frameworks.

    Graph fusion can combine operations into a single op to obtain higher accuracy and
        performance. Pattern like: Conv + Relu can be reduced to ConvRelu. This fusion
        will reduce memory accesses, and the quantization point after conv can also be
        removed.

    Technically we can fuse those layers before quantization, while fused layers are
        not supported by onnx standard. So to say ConvRelu is not a valid onnx
        operation, no execution framework can parse it.

    Therefore, PPQ will simulate the graph fusion by adjusting quantization config:
        if PPQ finds their is a pattern like Conv + Relu, the output quantization of
        Conv will be disabled, pretending that the Conv + Relu fusion has happened.

    This Pass is designed for 2 types graph fusion:
        1. activation fusion
        2. passive operation fusion

    For activation fusion, PPQ will identify the pattern: Computing op + Activation Op
        from your network. The output quantization of computing op will be disabled
        with their state being set to QuantizationState.OVERLAPPED.

    Activation fusion here supports only simple activation patterns,
        for complex activation functions like mish, swish, will be represented as
        mish = tanh + mul + softplus, swish = sigmoid + mul in onnx, cause onnx does
        not have a op definition for them. Identifying those complex patterns requires
        pattern matching, which is implemented in ppq.IR.search.py

    Complex quantization fusions must be invoked manually, PPQ implemented softplus &
        swish fusion functions in:
        :class:`mppq.quantization.optim.refine.MishFusionPass`
        :class:`mppq.quantization.optim.refine.SwishFusionPass`

    For passive operation fusion, PPQ will keep the input and the output variable share
        a same scale for passive operations. An operation is identified as passive op
        only if its attribute "is_active_quant_op" = False, this attribute is
        initialized by quantizer.

    If there is a passive operation having multiple input and output, the fusion
    procedure will make its FIRST input variable and ALL output variables share the
    same scale(the same scale as its first input). The quantization states of all
    output variables will be set to QuantizationState.OVERLAPPED.

    ### Parameters:

    * activation_type(Set[str]):

            A collection contains all activation types.

            The pattern will be recognized as [Computing Op -> Activation Op],

            By graph fusion, the output quantization of the Computing Op and
                the input quantization of the activation op will be disabled.

    * fuse_activation(bool)

            Whether to fuse activation op with computing op.

    # fuse_passive_op(bool)

            Whether to fuse passive op so that the input variable and output variable
            will share a same scale.

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this
    optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True

        # calling ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(
        self,
        activation_type: Set[str] = {"Relu", "LeakyRelu", "Clip"},
        fuse_activation: bool = True,
        fuse_passive_op: bool = True,
        fuse_relu_clip: bool = True,
    ) -> None:
        self.fuse_activation = fuse_activation
        self.fuse_passive_op = fuse_passive_op
        self.fuse_relu_clip = fuse_relu_clip
        self.activation_types = activation_type
        super().__init__(name="PPQ Quantization Fusion Pass")

    def is_same_platform(self, operations: List[Operation]):
        platforms = [operation.precision for operation in operations]
        return all([platform == platforms[0] for platform in platforms])

    @empty_ppq_cache
    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> None:
        processor = SearchableGraph(graph)

        # fuse computing operations and its following activation.
        if self.fuse_activation:
            patterns = processor.pattern_matching(
                patterns=[
                    lambda op: op.is_computing_op,
                    lambda op: op.type in self.activation_types,
                ],
                edges=[[0, 1]],
                exclusive=True,
            )

            for computing_op, act_op in patterns:
                if not isinstance(act_op, QuantableOperation):
                    continue
                if not isinstance(computing_op, QuantableOperation):
                    continue

                if (
                    computing_op.precision != act_op.precision
                    and computing_op.config.output_quantization_config[0].state
                    != QuantizationStates.FP32
                ):
                    warning(
                        f"Unexpected dispatching was found: "
                        f"Op {computing_op.name} and {act_op.name} should be send "
                        "to a same platform."
                    )
                    continue

                if (
                    len(graph.get_downstream_operations(computing_op)) == 1
                    and len(graph.get_upstream_operations(act_op)) == 1
                ):
                    computing_op.config.output_quantization_config[0].dominated_by = (
                        act_op.config.output_quantization_config[0]
                    )
                    act_op.config.input_quantization_config[0].dominated_by = (
                        act_op.config.output_quantization_config[0]
                    )

            if "Swish" in self.activation_types:
                search_engine = SearchableGraph(graph)
                patterns = search_engine.pattern_matching(
                    patterns=[lambda op: op.is_computing_op, "Sigmoid", "Mul"],
                    edges=[[0, 1], [1, 2], [0, 2]],
                    exclusive=True,
                )

                for pattern in patterns:
                    if any(
                        [not isinstance(op, QuantableOperation) for op in pattern],
                    ) or any(
                        [op.precision != pattern[0].precision for op in pattern],
                    ):
                        warning(
                            "There is a pattern of swish activation in your network "
                            f"start from {pattern[0]}, however part of your swish "
                            "activation is not quantable, so that graph fusion can not "
                            "merge their quantization configuration."
                        )
                        continue
                    computing, sigmoid, mul = pattern

                    assert isinstance(computing, QuantableOperation)
                    assert isinstance(sigmoid, QuantableOperation)
                    assert isinstance(mul, QuantableOperation)

                    master_config = mul.config.output_quantization_config[0]
                    computing.config.output_quantization_config[0].dominated_by = (
                        master_config
                    )
                    sigmoid.config.input_quantization_config[0].dominated_by = (
                        master_config
                    )
                    sigmoid.config.output_quantization_config[0].dominated_by = (
                        master_config
                    )
                    mul.config.input_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[1].dominated_by = master_config

            if "Mish" in self.activation_types:
                search_engine = SearchableGraph(graph)
                patterns = search_engine.pattern_matching(
                    patterns=[lambda op: op.is_computing_op, "Softplus", "Tanh", "Mul"],
                    edges=[[0, 1], [1, 2], [2, 3], [0, 3]],
                    exclusive=True,
                )

                for pattern in patterns:
                    if any(
                        [not isinstance(op, QuantableOperation) for op in pattern],
                    ) or any(
                        [op.precision != pattern[0].precision for op in pattern],
                    ):
                        warning(
                            "There is a pattern of mish activation in your network "
                            f"start from {pattern[0]}, however part of your mish "
                            "activation is not quantable, so that graph fusion can not "
                            "merge their quantization configuration."
                        )
                        continue
                    computing, softplus, tanh, mul = pattern

                    assert isinstance(computing, QuantableOperation)
                    assert isinstance(softplus, QuantableOperation)
                    assert isinstance(tanh, QuantableOperation)
                    assert isinstance(mul, QuantableOperation)

                    master_config = mul.config.output_quantization_config[0]
                    computing.config.output_quantization_config[0].dominated_by = (
                        master_config
                    )
                    tanh.config.input_quantization_config[0].dominated_by = (
                        master_config
                    )
                    tanh.config.output_quantization_config[0].dominated_by = (
                        master_config
                    )
                    softplus.config.input_quantization_config[0].dominated_by = (
                        master_config
                    )
                    softplus.config.output_quantization_config[0].dominated_by = (
                        master_config
                    )
                    mul.config.input_quantization_config[0].dominated_by = master_config
                    mul.config.input_quantization_config[1].dominated_by = master_config

        if self.fuse_passive_op:
            # all passive operations should never changes quantization configuration of
            # its input so to say their input and output share a same scale.
            for op in graph.operations.values():
                if op.type not in PASSIVE_OPERATIONS:
                    continue
                source_op = op.inputs[0].source_op
                if source_op is None:
                    continue  # beginning op, can not merge.
                if isinstance(op, QuantableOperation) and self.is_same_platform(
                    [op, source_op]
                ):
                    TQC = op.config.input_quantization_config[0]
                    for output_cfg in op.config.output_quantization_config:
                        output_cfg.dominated_by = TQC

        if self.fuse_relu_clip:
            patterns = processor.pattern_matching(
                patterns=[lambda op: True, lambda op: op.type in {"Relu", "Clip"}],
                edges=[[0, 1]],
                exclusive=True,
            )
            for computing_op, act_op in patterns:
                if not isinstance(act_op, QuantableOperation):
                    continue
                if not isinstance(computing_op, QuantableOperation):
                    continue

                if (
                    len(graph.get_downstream_operations(computing_op)) == 1
                    and len(graph.get_upstream_operations(act_op)) == 1
                ):
                    computing_op.config.output_quantization_config[0].dominated_by = (
                        act_op.config.output_quantization_config[0]
                    )
                    act_op.config.input_quantization_config[0].dominated_by = (
                        act_op.config.output_quantization_config[0]
                    )


@OPTIM_ALGORITHMS.register()
class QuantAlignmentPass(QuantizationOptimizationPass):
    r"""PPQ Quant Alignment Pass(通用量化对齐过程)

    When deploy on real hardware and inference framework, we will find that there are
    various restrictions or rules that we have to follow.

    * AVERAGE_POOL_2D: Input and outputs must all have same scale/zero_point

    * CONCATENATION: Input and outputs must all have same scale/zero_point

    * SLICE: Input and outputs must all have same scale/zero_point

    More detailed restrictions please refer to:
    https://www.tensorflow.org/lite/performance/quantization_spec

    Those restrictions, can be concluded as some quantization should share
        the same quantization parameter with others. PPQ Quant Alignment Pass is
        designed for dealing with problems like this.

    PPQ uses Tensor Quantization Config (A data structure defined in ppq.core) to
        control the quantization logic, so to say if we want to align quantization
        parameters, we align their TQC in fact.

    The way to align TQC is simple, code like:
        tqc1.set_master(master=tqc2)
    Will make tqc1 and tqc2 share the same quantization parameters as tqc1 has, and
    change the state of tqc2 to be QuantizationState.PASSIVE

    If we access the scale of tqc2, PPQ will return its master TQC's scale instead,
    so does offset.

    That is tqc1 and tqc2 are bonuded with statement "tqc1.set_master(master=tqc2)".

    ### Parameters:

    * elementwise_alignment(Set[str]):

            Alignment method for elementwise ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All elementwise ops are listed in ppq.core.common.py

    * concat_alignment(Set[str])

            Alignment method for concat-like ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All concat-like ops are listed in ppq.core.common.py

    * averagepool_alignment(Set[str])

            Alignment method for pooling-like ops.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

            All pooling-like ops are listed in ppq.core.common.py

    * resize_alignment(Set[str])

            Alignment method for Resize op.

            PPQ Supports 4 alignment methods:
                namely 'Align to Input', 'Align to Large', 'Align to Output', 'None'.

    * force_overlap(bool)

            TQC alignment might cause serious cascade effect.

            For subgraph like this:

            Conv1 ---
                    + --- Add1
            Conv2 ---
                    + --- Conv3

            If we demand Add1 to have same input scale, this alignment will affect
                Conv3 also, cause Conv2's output is feed to both Add1 and Conv3.

            If force_overlap = False, PPQ alignment procedure will remain the output
                scale of Conv2 as unchanged, while only align the input scale of Add1.

            If force_overlap = True, the input of Add1, Conv3 and the output of Conv2
                will all be aligned to a same scale.

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this
    optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.fusion = True
        setting.fusion_setting.force_alignment_overlap = True

        # calling ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(
        self,
        elementwise_alignment: Literal[
            "Align to Large", "Align to Output", "None"
        ] = "Align to Large",
        concat_alignment: Literal[
            "Align to Large", "Align to Output", "None"
        ] = "Align to Output",
        pooling_alignment: Literal[
            "Align to Input", "Align to Output", "None"
        ] = "None",
        resize_alignment: Literal[
            "Align to Input", "Align to Output", "None"
        ] = "Align to Output",
        force_overlap: bool = False,
    ) -> None:
        self.pooling_alignment = pooling_alignment
        self.elementwise_alignment = elementwise_alignment
        self.concat_alignment = concat_alignment
        self.resize_alignment = resize_alignment
        self.force_overlap = force_overlap
        assert self.elementwise_alignment in {
            "Align to Large",
            "Align to Output",
            "None",
        }
        assert self.concat_alignment in {"Align to Large", "Align to Output", "None"}
        assert self.pooling_alignment in {"Align to Input", "Align to Output", "None"}
        assert self.resize_alignment in {"Align to Input", "Align to Output", "None"}
        super().__init__(name="PPQ Quantization Alignment Pass")

    def align_to_input(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to input config. All output configs
        would share a same scale and offset with output config. (as a slave to
        input config)

        Any change to slave config will be rejected since then.
        """
        master_config = op.config.input_quantization_config[0]
        for slave_config in op.config.output_quantization_config:
            if slave_config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            elif slave_config.state != QuantizationStates.FP32:
                slave_config.master_by = master_config
        return master_config

    def align_to_large(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to larger input config. The first input
        config will be set as master config, all slave config will share the
        same scale and offset with master.

        Any change to slave config will be rejected since then.
        """
        global_min, global_max, master_config = (
            0,
            0,
            op.config.input_quantization_config[0],
        )
        for config in op.config.input_quantization_config:
            if config.state == QuantizationStates.FP32:
                continue
            elif config.policy.has_property(QuantizationProperty.FLOATING):
                continue

            assert config.policy.has_property(
                QuantizationProperty.PER_TENSOR
            ), "Quant Alignment can only happen with per tensor quantization."
            local_min = config.scale * (config.quant_min - config.offset)
            local_max = config.scale * (config.quant_max - config.offset)

            assert isinstance(local_min, torch.Tensor)
            assert isinstance(local_max, torch.Tensor)
            global_max = max(global_max, local_max.item())
            global_min = min(global_min, local_min.item())

        # recompute scale and offset
        scale, offset = minmax_to_scale_offset(
            global_min, global_max, op.config.input_quantization_config[0]
        )

        device = master_config.scale.device
        master_config._dominator = master_config
        master_config.state = QuantizationStates.PASSIVE
        master_config.scale = torch.tensor(scale, dtype=torch.float32, device=device)
        master_config.offset = torch.tensor(offset, dtype=torch.float32, device=device)

        for slave_config in op.config.input_quantization_config[1:]:
            if slave_config.state == QuantizationStates.FP32:
                continue
            elif slave_config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            slave_config.master_by = master_config
        return master_config

    def align_to_output(self, op: QuantableOperation) -> TensorQuantizationConfig:
        """Align quant scale and offset to output config. All input configs
        would share a same scale and offset with output config. (as a slave to
        output config)

        Any change to slave config will be rejected since then.
        """
        master_config = op.config.output_quantization_config[0]
        for slave_config in op.config.input_quantization_config:
            if slave_config.policy.has_property(QuantizationProperty.FLOATING):
                continue
            elif slave_config.state != QuantizationStates.FP32:
                slave_config.master_by = master_config
        return master_config

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> None:
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation):
                continue

            master_config = None
            if operation.type in TYPES_FOR_ALIGNMENT["Elementwise"]:
                if self.elementwise_alignment == "None":
                    continue
                if self.elementwise_alignment == "Align to Large":
                    master_config = self.align_to_large(operation)
                else:
                    master_config = self.align_to_output(operation)

            elif operation.type in TYPES_FOR_ALIGNMENT["Concat"]:
                if self.concat_alignment == "None":
                    continue
                if self.concat_alignment == "Align to Large":
                    master_config = self.align_to_large(operation)
                else:
                    master_config = self.align_to_output(operation)

            elif operation.type in TYPES_FOR_ALIGNMENT["Pooling"]:
                if self.pooling_alignment == "None":
                    continue
                if self.pooling_alignment == "Align to Input":
                    self.align_to_input(operation)  # do not set master_config
                if self.pooling_alignment == "Align to Output":
                    master_config = self.align_to_output(operation)

            elif operation.type == "Resize":
                if self.resize_alignment == "None":
                    continue
                if self.resize_alignment == "Align to Output":
                    master_config = self.align_to_output(operation)
                if self.resize_alignment == "Align to Input":
                    self.align_to_input(operation)  # do not set master_config

            elif ALIGNMENT_MANUL_OVERRIDE in operation.extension_attrib:
                method = operation.extension_attrib[ALIGNMENT_MANUL_OVERRIDE]
                if self.concat_alignment == "Align to Large":
                    master_config = self.align_to_large(operation)
                elif self.concat_alignment == "Align to Large":
                    master_config = self.align_to_output(operation)
                else:
                    warning(
                        f"Unrecognized Alignment Method {method} for "
                        f"operation {operation.name}"
                    )

            if master_config is not None:
                # override up stream layer's config if possible
                for up_op in graph.get_upstream_operations(operation):
                    if not isinstance(up_op, QuantableOperation):
                        continue

                    if (
                        len(graph.get_downstream_operations(up_op)) != 1
                        and not self.force_overlap
                    ):
                        continue
                    for cfg, var in up_op.config_with_variable:
                        if operation in var.dest_ops:
                            cfg.master_by = master_config


@OPTIM_ALGORITHMS.register()
class SwishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("Swish Fusion")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda op: op.is_computing_op, "Sigmoid", "Mul"],
            edges=[[0, 1], [1, 2], [0, 2]],
            exclusive=True,
        )

        for pattern in patterns:
            if any(
                [not isinstance(op, QuantableOperation) for op in pattern],
            ) or any(
                [op.precision != pattern[0].precision for op in pattern],
            ):
                warning(
                    "There is a pattern of swish activation in your network start "
                    f"from {pattern[0]}, however part of your swish activation is "
                    "not quantable, so that graph fusion can not merge their "
                    "quantization configuration."
                )
                continue
            computing, sigmoid, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(sigmoid, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            master_config = mul.config.output_quantization_config[0]
            computing.config.output_quantization_config[0].dominated_by = master_config
            sigmoid.config.input_quantization_config[0].dominated_by = master_config
            sigmoid.config.output_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[1].dominated_by = master_config


@OPTIM_ALGORITHMS.register()
class MishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("Mish Fusion")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda op: op.is_computing_op, "Softplus", "Tanh", "Mul"],
            edges=[[0, 1], [1, 2], [2, 3], [0, 3]],
            exclusive=True,
        )

        for pattern in patterns:
            if any(
                [not isinstance(op, QuantableOperation) for op in pattern],
            ) or any(
                [op.precision != pattern[0].precision for op in pattern],
            ):
                warning(
                    "There is a pattern of mish activation in your network start "
                    f"from {pattern[0]}, however part of your mish activation is not "
                    "quantable, so that graph fusion can not merge their quantization "
                    "configuration."
                )
                continue
            computing, softplus, tanh, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(softplus, QuantableOperation)
            assert isinstance(tanh, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            master_config = mul.config.output_quantization_config[0]
            computing.config.output_quantization_config[0].dominated_by = master_config
            tanh.config.input_quantization_config[0].dominated_by = master_config
            tanh.config.output_quantization_config[0].dominated_by = master_config
            softplus.config.input_quantization_config[0].dominated_by = master_config
            softplus.config.output_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[0].dominated_by = master_config
            mul.config.input_quantization_config[1].dominated_by = master_config



"""
This File Contains Legacy Dispatchers.
Refer to ppq.scheduler.perseus for updated implementation.
"""

from typing import Collection, Dict

from mppq.dispatcher.base import DISPATCHER_TABLE
from mppq.dispatcher.conservative import ConservativeDispatcher
from mppq.quant import TargetPrecision


@DISPATCHER_TABLE.register("pointwise")
class PointDispatcher(ConservativeDispatcher):
    r"""Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable
    operations, Shape-Or-Index related operations and remaining operations, all sub
    classes of GraphDispatcher will give an implementation of function "dispatch" to
    send all operations to their proper platform.

    Point Dispatch send all computing op to quantable platform, while other ops remain
    unchanged.

    ATTENTION:

        platform attribute will greatly affect quantizer's quantization logic, and the
        execution result.

        - If operation is sent to a quantable platform, then its inputs and outputs
          will be quantized if necessary.
        - If operation is classified as shape-or-index related operation, then its
          execution will be taken with cpu.
        - If operation is sent to a fp32 platform, then its inputs and outputs shall
          never be quantized.

    ATTENTION:

        this dispatcher will insert necessary DeviceSwitch operations between
        shape-or-index operations and others.
    """

    def dispatch(
        self,
        quant_types: Collection[str],
        quant_precision: TargetPrecision,
        fp32_platform: TargetPrecision = TargetPrecision.FP32,
        soi_platform: TargetPrecision = TargetPrecision.SOI,
        **kwargs,
    ) -> Dict[str, TargetPrecision]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_precision - all quantable parts of graph will be dispatched to this
                             platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched
                             to this platform.
            fp32_platform  - there are some operations receiving results from both
                             quant_precision and SOI_platform, they will be dispatched
                             they will be dispatched to fp32_platform.

        ATTENTION:

            Quantization follows this dispatching, and only the operations within
            quantable platform will be quantized in the future.

        ATTENTION:

            this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this
                dispatcher.
            quant_types(Set[str]): all quantable types for given platforms.
            quant_precision (TargetPlatform): =
                platform object where quantable parts will goes to.
            SOI_platform (TargetPlatform):
                platform object where SOI parts will goes to.
            fp32_platform (TargetPlatform):
                platform object where remaining parts will goes to.
        """
        graph = self.graph

        dispatch_table = super().dispatch(
            quant_types=quant_types,
            quant_precision=quant_precision,
            fp32_platform=fp32_platform,
            soi_platform=soi_platform,
            kwargs=kwargs,
        )

        skip_ops = set()
        for op in graph.operations.values():
            if op in skip_ops:
                continue
            if op.type in quant_types and op.is_computing_op:
                dispatch_table[op.name] = quant_precision
            else:
                if dispatch_table[op.name] == quant_precision:
                    dispatch_table[op.name] = fp32_platform

        return dispatch_table



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Sequence, Type

import torch

from mppq.executor.base import QuantRuntimeHook
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.quantize import QuantableOperation
from mppq.logger import error, info
from mppq.quant import QuantizationStates, TensorQuantizationConfig
from mppq.register import Registry


class BaseTensorObserver(metaclass=ABCMeta):
    """A base class for all tensor observers."""

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        self._watch_on = watch_on
        self._quant_cfg = quant_cfg

    @abstractmethod
    def observe(self, value: Any):
        raise NotImplementedError("Implement this function first.")

    @abstractmethod
    def render_quantization_config(self):
        raise NotImplementedError("Implement this function first.")

    def __repr__(self) -> str:
        return (
            "PPQ Tensor Observer ("
            + self.__class__.__name__
            + ") mount on variable "
            + self._watch_on.name
            + " observing algorithm: "
            + self._quant_cfg.observer_algorithm
        )

    def report(self) -> str:
        if self._quant_cfg.state == QuantizationStates.ACTIVATED:
            return (
                f"Observer on Variable {self._watch_on.name}, "
                f"computed scale: {self._quant_cfg.scale}, "
                f"computed offset: {self._quant_cfg.offset}\n"
            )
        return ""


OBSERVER_TABLE: Registry[Type[BaseTensorObserver]] = Registry("OBSERVER_TABLE")


class CalibrationHook(QuantRuntimeHook):
    """A runtime hook for operation observer."""

    def __init__(
        self,
        operation: QuantableOperation,
        observer_table: Dict[Variable, BaseTensorObserver],
    ) -> None:
        self._operation = operation
        self._observer_table = observer_table
        super().__init__(operation)

    def pre_forward_hook(
        self,
        inputs: Sequence[torch.Tensor],
        quant_inputs: Sequence[torch.Tensor] = (),
        quant_configs: Sequence[TensorQuantizationConfig] = (),
        **kwargs,
    ):
        for input_var, quant_config in zip(inputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(input_var)
        return super().pre_forward_hook(inputs, quant_inputs, quant_configs)

    def post_forward_hook(
        self,
        outputs: Sequence[torch.Tensor],
        quant_outputs: Sequence[torch.Tensor] = (),
        quant_configs: Sequence[TensorQuantizationConfig] = (),
        **kwargs,
    ):
        for output_var, quant_config in zip(outputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(output_var)
        return super().post_forward_hook(outputs, quant_outputs, quant_configs)

    def render_quantization_config(self):
        for _, observer in self._observer_table.items():
            observer.render_quantization_config()
            observer.report()

    def __repr__(self) -> str:
        return "".join(
            [observer.__str__() + "\n" for _, observer in self._observer_table.items()]
        )


class OperationObserver:
    def __init__(
        self,
        operation: Operation,
        monitor_parameter: bool = True,
        monitor_outputs: bool = True,
        monitor_inputs: bool = True,
    ) -> None:
        self._operation = operation
        self._hook = OperationObserver.build_hook(
            operation=operation,
            monitor_parameter=monitor_parameter,
            monitor_outputs=monitor_outputs,
            monitor_inputs=monitor_inputs,
        )

    @staticmethod
    def build_observer(
        variable: Variable, config: TensorQuantizationConfig
    ) -> BaseTensorObserver:
        algorithm = str(config.observer_algorithm.lower())
        if algorithm not in OBSERVER_TABLE:
            error(f"observer algorithm {algorithm} not found")
            info(f"{OBSERVER_TABLE}")
            raise KeyError
        return OBSERVER_TABLE[algorithm](watch_on=variable, quant_cfg=config)

    @staticmethod
    def build_hook(
        operation: Operation,
        monitor_parameter: bool,
        monitor_outputs: bool,
        monitor_inputs: bool,
    ) -> CalibrationHook:
        if not isinstance(operation, QuantableOperation):
            raise TypeError(
                f"Only QuantableOP instance can apply an Observer, "
                f"while {type(operation)} was given."
            )
        observer_table = {}
        for var, cfg in zip(
            operation.inputs, operation.config.input_quantization_config
        ):
            if cfg.state == QuantizationStates.INITIAL:
                if var.is_parameter and monitor_parameter:
                    observer_table[cfg] = OperationObserver.build_observer(var, cfg)
                elif not var.is_parameter and monitor_inputs:
                    observer_table[cfg] = OperationObserver.build_observer(var, cfg)

        if monitor_outputs:
            for var, cfg in zip(
                operation.outputs,
                operation.config.output_quantization_config,
            ):
                if cfg.state == QuantizationStates.INITIAL:
                    observer_table[cfg] = OperationObserver.build_observer(var, cfg)

        return CalibrationHook(operation=operation, observer_table=observer_table)

    def render_quantization_config(self):
        self._hook.render_quantization_config()

    @property
    def hook(self) -> CalibrationHook:
        return self._hook

    def report(self) -> str:
        """print debug messages"""
        return str(self._hook)



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

import re
from dataclasses import dataclass, field
from functools import partial
from typing import Dict, List, Optional

from mppq.ir.base.graph import BaseGraph
from mppq.ir.search import SearchableGraph
from mppq.logger import nest
from mppq.quant import TargetPrecision


@dataclass
class Subgraph:
    """Define a subgraph in ``IgnoredScope``."""

    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)


class IgnoredScope:
    r"""Create a scope to depict ignored quantizing operations in the graph.

    Args:
        types (List[str], optional): The types of operations to be ignored.
            Defaults to None.
        subgraphs (List[Subgraph | dict], optional): A subgraph which is
            defined by a list of input operations and output operations.
            Defaults to None.
        operations (List[str], optional): The names of operations to be ignored.
            The name can be either a fully matched string, a fuzzy string including
            wildcard characters ('*' and '?'), or a regular expression pattern
            which starts with "re://". Defaults to None.

    Example::

        model = dict(
            ignore=dict(
                type="IgnoredScope",
                types=["Conv", "Gemm"],
                operations=[r"re://.*conv.*", "Gemm?", "Concat_22"],
                subgraphs=[
                    dict(type="Subgraph", inputs=["input"], outputs=["output"])
                ],
            )
        )

    Note:

        Character "?" is a legal character in onnx name, if the op name contains '?', it
        will not be accurately matched, unless use the regular expression pattern.
        For example, to match the op name "Conv_123?split0", use the following string:
        r"re://Conv_123\?split0".
    """

    CASE_INSENSITIVE = False

    def __init__(
        self,
        types: Optional[List[str]] = None,
        subgraphs: Optional[List[Subgraph | dict]] = None,
        operations: Optional[List[str | re.Pattern]] = None,
    ):
        self.types = types or []
        self.subgraphs = subgraphs or []
        for i, subgraph in enumerate(self.subgraphs):
            if isinstance(subgraph, dict):
                self.subgraphs[i] = Subgraph(
                    inputs=subgraph.get("inputs", []),
                    outputs=subgraph.get("outputs", []),
                )
        self.operations = operations or []
        # normalize to regular expression
        for i, op in enumerate(self.operations):
            if isinstance(op, re.Pattern):
                continue
            if op.startswith("re://"):
                self.operations[i] = re.compile(op[5:])
            elif "*" in op or "?" in op:
                self.operations[i] = re.compile(op.replace("*", ".*").replace("?", "."))
        if IgnoredScope.CASE_INSENSITIVE and types:
            self.types = [t.lower() for t in types]
        self.log = nest(self.__class__.__name__)

    def dispatch(
        self, graph: BaseGraph, platform: TargetPrecision = TargetPrecision.FP32
    ) -> Dict[str, TargetPrecision]:
        """Dispatch the ignored scope to the graph.

        Args:
            graph (BaseGraph): The graph to apply the ignored scope to.
            platform (TargetPlatform, optional): Specify the platform for ignored
                operations. Defaults to TargetPlatform.FP32.
        """

        dispatch_table = {}
        for operation in graph.operations.values():
            op_type = operation.type
            if IgnoredScope.CASE_INSENSITIVE:
                op_type = operation.type.lower()
            op_name = operation.name
            if IgnoredScope.CASE_INSENSITIVE:
                op_name = operation.name.lower()
            if self.types and op_type in self.types:
                dispatch_table[operation.name] = platform
            if op_name in self.operations:
                dispatch_table[operation.name] = platform
            for p in filter(lambda x: isinstance(x, re.Pattern), self.operations):
                assert isinstance(p, re.Pattern)
                if p.fullmatch(op_name) is not None:
                    dispatch_table[operation.name] = platform

        def _op_match(op, pattern):
            return op.name in pattern

        for subgraph in self.subgraphs:
            assert isinstance(subgraph, Subgraph)
            operations = SearchableGraph(graph).opset_matching(
                sp_expr=partial(_op_match, pattern=subgraph.inputs),
                ep_expr=partial(_op_match, pattern=subgraph.outputs),
                rp_expr=lambda x, y: True,
                direction="down",
            )
            for operation in operations:
                dispatch_table[operation.name] = platform

        for operation in dispatch_table:
            self.log.debug(f"Ignored operations: {operation}")
        return dispatch_table



# PPQ 全局配置，你可以自由修改下列属性以达成特定目的。
# PPQ System configuration
# You can modify following codes for your own purpose.

# Observer 中，最小 scale 限制，所有小于该值的 scale 将被该值覆盖
OBSERVER_MIN_SCALE = 1e-4
# Observer 中，最小 scale 的手动覆盖属性
OBSERVER_MIN_SCALE_MANUL_OVERRIDE = "OBSERVER_MIN_SCALE_MANUL_OVERRIDE"
# Observer 中 kl 散度的计算设备
OBSERVER_KL_COMPUTING_DEVICE = "cpu"
# Observer 中 hist 箱子的个数
OBSERVER_KL_HIST_BINS = 4096
# Observer 中 hist 箱子的属性
OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE = "OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE"
# Observer 中 percentile 的参数
OBSERVER_PERCENTILE = 0.9999
# Observer 中 percentile 参数的属性
OBSERVER_PERCENTILE_MANUL_OVERRIDE = "OBSERVER_PERCENTILE_MANUL_OVERRIDE"
# Observer 中 mse 校准方法 hist 箱子的个数
OBSERVER_MSE_HIST_BINS = 2048
# Observer 中 mse 计算的间隔，间隔越小，所需时间越长
OBSERVER_MSE_COMPUTE_INTERVAL = 8
# Floating MSE Observer 的样本个数
OBSERVER_FLOATING_MSE_FETCHES = 4096
# Isotone Observer 的监听轴
# 对于 Softmax 激活函数而言，Isotone Observer 的监听轴应该与 Softmax 操作所规约的轴相同
# 对于 Sigmoid 激活函数而言，Isotone Observer 的监听轴应该设置为 Batch 所在的轴
OBSERVER_ISOTONE_OBSERVER_AXIS = "OBSERVER_ISOTONE_OBSERVER_AXIS"

# PASSIVE OPERATIONS 是那些不参与计算的 Op, 这些 op 的输入与输出将直接共享 scale
# 同时这些 op 前后的定点过程将被直接停用
PASSIVE_OPERATIONS = {
    "MaxPool",
    "GlobalMaxPool",
    "Reshape",
    "Flatten",
    "Identity",
    "Dropout",
    "Slice",
    "Pad",
    "Split",
    "Transpose",
    "Interp",
    "Squeeze",
    "Unsqueeze",
    "Resize",
}
# COPUTING OP 是所有带参数的计算层，该属性被用于联合定点和子图切分
COMPUTING_OP = {
    "Conv",
    "Gemm",
    "ConvTranspose",
    "MatMul",
    "Attention",
}
# Activation 是所有激活层，该属性被用于联合定点和子图切分
ACTIVATION_OP = {
    "Relu",
    "LeakyRelu",
    "Sigmoid",
    "Tanh",
    "Clip",
    "PRelu",
    "Elu",
    "HardSigmoid",
    "HardSwish",
    "SiLU",
    "Mish",
    "Swish",
    "Gelu",
}
# 默认量化的算子类型
DEFAULT_QUANTIZE_OP = PASSIVE_OPERATIONS | COMPUTING_OP | ACTIVATION_OP

# 强制联合定点的算子种类
TYPES_FOR_ALIGNMENT = {
    "Elementwise": {"Add", "Sub", "Sum"},
    "Concat": {"Concat"},
    "Pooling": {"AveragePool", "GlobalAveragePool"},
}
# 强制联合定点手动覆盖
ALIGNMENT_MANUL_OVERRIDE = "ALIGNMENT_MANUL_OVERRIDE"

# ONNX 导出图的时候，opset的版本，这玩意改了可能就要起飞了
ONNX_EXPORT_OPSET = 19
# ONNX 导出图的时候，onnx version，这玩意改了可能就要起飞了
ONNX_VERSION = 9
ONNX_DOMAIN = "ai.onnx"

DEFAULT_OPSET_DOMAIN = ONNX_DOMAIN
DEFAULT_OPSET_VERSION = ONNX_EXPORT_OPSET
STRICT_OPSET_CHECKING = False

# LSTM 算子的权重缓存属性
LSTM_FLATTEN_WEIGHT_ATTRIB = "LSTM_FLATTEN_WEIGHT_ATTRIB"
# GRU 算子的权重缓存属性
GRU_FLATTEN_WEIGHT_ATTRIB = "GRU_FLATTEN_WEIGHT_ATTRIB"
# 图上用于表示 Opset 的属性
GRAPH_OPSET_ATTRIB = "GRAPH_OPSET"

# LINEAR ACTIVATIONS 是所有线性激活层，PPQ 将执行计算层与线性激活层的联合定点，不论后端是否真的做了图融合。
# 事实上就算后端不融合这些层，执行联合定点也是有益无害的。
LINEAR_ACTIVATIONS = {"Relu", "Clip"}

# 误差容忍度
CHECKPOINT_TOLERANCE = 1

# 要做 Bias Correction 的算子种类
BIAS_CORRECTION_INTERST_TYPE = {"Conv", "Gemm", "ConvTranspose"}

# 导出 qdq 节点时是否需要导出状态已经是 overlap 的节点
EXPORT_OVERLAPPED_CONFIG = False

# 可以产生SOI数据的算子
SOI_DATA_GENERATOR_OPS = {
    "Constant",
    "ConstantOfShape",
    "NonMaxSuppression",
    "Shape",
    "TopK",
}

# 支持整型计算的算子。可以用来传播SOI。
INT_CALCULATION_OPS = {
    "Reshape",
    "Slice",
    "Gather",
    "Pad",
    "Resize",
    "Split",
    "TopK",
    "Tile",
    "Expand",
    "RoiAlign",
    "MMCVRoiAlign",
}



import torch

from mppq.ir.base.quantize import BaseQuantFunction
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from mppq.utils.qfunction.floating import floating_quant
from mppq.utils.qfunction.linear import (
    dynamic_linear_quant,
    linear_fake_quant,
    linear_quant_toint,
)


def ppq_fake_quant(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，量化给定的 tensor.

    请注意 config.state 必须处在激活状态，该函数起作用。如果 config.state 处于
        INITIAL, FP32, PASSIVE_INIT 等未激活状态，该函数不进行任何处理，直接返回 tensor.

    ### 线性量化(QuantizationProperty.LINEAR):

        INT8 = Clip(Round((FP32 / scale)))

    ### 浮点量化(QuantizationProperty.FLOATING):

        FP8 = Clip(FP32_TO_FP8((FP32 / scale)))

    ### 动态线性量化(QuantizationProperty.DYNAMIC)

        scale = max(FP32) / 255

        INT8  = Clip(Round((FP32 / scale)))

    """
    if tensor is None:
        raise ValueError("Tensor is empty.")
    if not QuantizationStates.is_activated(config.state):
        return tensor
    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return linear_fake_quant(tensor, config)
        else:
            return dynamic_linear_quant(tensor, config)

    if config.policy.has_property(QuantizationProperty.FLOATING):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return floating_quant(tensor, config)

    raise ValueError(
        "Unexpected Quantization Property Found in ppq_fake_quant. "
        "Do not know how to quantize your config yet."
    )


def ppq_quant_toint(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，这个函数将会执行线性量化，动态量化

    但是结果直接出来是整数
    """
    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return linear_quant_toint(tensor, config)

    raise ValueError(
        "Unexpected Quantization Property Found in ppq_quant_toint. "
        "Do not know how to quantize your config yet."
    )


__all__ = [
    "ppq_fake_quant",
    "ppq_quant_toint",
    "BaseQuantFunction",
]



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import (
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    overload,
)

import torch
from torch import Tensor

from mppq.executor.op.base import DEFAULT_BACKEND_TABLE, TorchBackendContext
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation
from mppq.ir.base.quantize import QuantableOperation
from mppq.quant import BuiltinPlatform, TensorQuantizationConfig

OPERATION_FORWARD_TABLE = {
    platform.value: deepcopy(DEFAULT_BACKEND_TABLE) for platform in BuiltinPlatform
}

GraphInput = TypeVar("GraphInput", Sequence[Tensor], Mapping[str, Tensor], Tensor)


class RuntimeHook:
    """RuntimeHook is an abstract class designed for executor customizing.

    Args:
        metaclass ([type], optional): [description]. Defaults to ABCMeta.
    """

    def __init__(self, operation: Operation, **kwargs) -> None:
        self._hook_to = operation

    def pre_forward_hook(self, inputs: Sequence[Tensor], **kwargs) -> List[Tensor]:
        """user-customized pre-processing procedure of input data.

        Args:
            inputs (list): a list includes all input data.

        Returns:
            list: a list includes all input data(processed).
        """
        return list(inputs)

    def post_forward_hook(self, outputs: Sequence[Tensor], **kwargs) -> List[Tensor]:
        """user-customized post-processing procedure of output data.

        Args:
            inputs (list): a list includes all output data.

        Returns:
            list: a list includes all output data(processed).
        """
        return list(outputs)


class QuantRuntimeHook(RuntimeHook):
    """QuantRuntimeHook is an abstract class designed for executor customizing."""

    def __init__(self, operation: Operation, **kwargs) -> None:
        if not isinstance(operation, QuantableOperation):
            raise TypeError(
                "You are trying to bind a QuantRuntimeHook "
                f"to a non-quantized operation {operation}."
            )
        super().__init__(operation, **kwargs)

    def pre_forward_hook(
        self,
        inputs: Sequence[Tensor],
        quant_inputs: Sequence[Tensor] = (),
        quant_configs: Sequence[TensorQuantizationConfig] = (),
        **kwargs,
    ) -> List[Tensor]:
        assert len(inputs) == len(quant_inputs) == len(quant_configs)
        return list(quant_inputs)

    def post_forward_hook(
        self,
        outputs: Sequence[Tensor],
        quant_outputs: Sequence[Tensor] = (),
        quant_configs: Sequence[TensorQuantizationConfig] = (),
        **kwargs,
    ) -> List[Tensor]:
        assert len(outputs) == len(quant_outputs) == len(quant_configs)
        return list(quant_outputs)


class BaseGraphExecutor(metaclass=ABCMeta):
    """PPQ Base Graph Executor.

    Args:
        Callable ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ABCMeta.
    """

    def __init__(
        self, graph: BaseGraph, target_platform: int = BuiltinPlatform.UNSPECIFIED
    ) -> None:
        self.load_graph(graph=graph)
        self.target_platform = target_platform

    @property
    def graph(self) -> BaseGraph:
        return self._graph

    def load_graph(self, graph: BaseGraph):
        self._graph = graph
        self._executing_order = self._graph.topological_sort()

    def _prepare_input(self, inputs: Optional[GraphInput]) -> Dict[str, Tensor]:
        inputs_dictionary = self._graph.inputs
        if len(inputs_dictionary) == 0:
            assert (
                inputs is None
            ), "Graph do not need any inputs. please set your inputs to be None."
            return {}

        if isinstance(inputs, torch.Tensor):
            assert (
                len(inputs_dictionary) == 1
            ), "Graph needs more than one input, while only one tensor was given."
            return {list(inputs_dictionary.keys())[0]: inputs}
        elif isinstance(inputs, Mapping):
            assert len(inputs_dictionary) == len(inputs), (
                f"Inputs format misunderstood. Given inputs has "
                f"{len(inputs)} elements, while graph needs {len(inputs_dictionary)}"
            )
            return {k: v for k, v in inputs.items()}
        elif isinstance(inputs, Sequence):
            assert len(inputs_dictionary) == len(inputs), (
                f"Inputs format misunderstood. Given inputs has "
                f"{len(inputs)} elements, while graph needs {len(inputs_dictionary)}"
            )
            return {k: v for k, v in zip(inputs_dictionary, inputs)}
        else:
            raise ValueError(f"Unsupported input type: {type(inputs)}")

    @abstractmethod
    def forward(
        self,
        inputs: GraphInput,
        output_names: Optional[List[str]] = None,
        hooks: Optional[Mapping[str, RuntimeHook]] = None,
    ) -> List[torch.Tensor]:
        """Forward a graph from given inputs to required output names.

        If output_names is not given, all graph outputs will be returned.
        hook is called before and after each operation's forward.
        """
        raise NotImplementedError

    @abstractmethod
    def tracing_operation_meta(
        self,
        inputs: GraphInput,
        output_names: Optional[List[str]] = None,
    ) -> None:
        raise NotImplementedError("Please implement this function first.")

    @overload
    def forward_single_operation(
        self,
        op: Operation,
        inputs: List[Tensor],
        ctx: Optional[TorchBackendContext] = None,
        return_list: Literal[True] = True,
    ) -> Tuple[Tensor, ...]:
        """"""

    @overload
    def forward_single_operation(
        self,
        op: Operation,
        inputs: List[Tensor],
        ctx: Optional[TorchBackendContext] = None,
        return_list: bool = True,
    ) -> Tuple[Tensor, ...] | Tensor:
        """"""

    def forward_single_operation(
        self,
        op: Operation,
        inputs: List[Tensor],
        ctx: Optional[TorchBackendContext] = None,
        return_list: bool = True,
    ) -> Tuple[Tensor, ...] | Tensor:
        """Forward a single operation with given inputs."""
        f = OPERATION_FORWARD_TABLE[self.target_platform][op.type]
        ret = f(op, inputs, ctx=ctx)
        if return_list and not isinstance(ret, (list, tuple)):
            return (ret,)
        return ret

    def __call__(
        self,
        inputs: GraphInput,
        output_names: Optional[List[str]] = None,
    ) -> List[torch.Tensor]:
        return self.forward(inputs=inputs, output_names=output_names)

    def __repr__(self) -> str:
        return f"PPQ GraphExecuter Object: {self.__hash__()}"
